from dataclasses import dataclass
from typing import Optional

from servey.security.authorization import Authorization

from persisty.factory.store_factory_abc import StoreFactoryABC
from persisty.store.store_abc import StoreABC
from persisty.store_meta import T, StoreMeta


@dataclass
class DefaultStoreFactory(StoreFactoryABC[T]):
    """
    No op store factory which doesn't actually use authorization
    """

    store: StoreABC

    def get_meta(self) -> StoreMeta:
        return self.store.get_meta()

    def create(self, authorization: Optional[Authorization]) -> Optional[StoreABC[T]]:
        return self.store



from copy import deepcopy
from time import time
from typing import Optional, Dict, List, Generic

from dataclasses import dataclass, field

from marshy import dump, get_default_context
from marshy.marshaller.marshaller_abc import MarshallerABC

from persisty.batch_edit import BatchEdit
from persisty.batch_edit_result import BatchEditResult
from persisty.result_set import ResultSet
from persisty.search_filter.include_all import INCLUDE_ALL
from persisty.search_filter.search_filter_abc import SearchFilterABC
from persisty.search_order.search_order import SearchOrder
from persisty.store.store_abc import StoreABC, T
from persisty.store_meta import StoreMeta
from persisty.util import secure_hash


@dataclass(frozen=True)
class TtlEntry(Generic[T]):
    value: T
    expire_at: int


@dataclass(frozen=True)
class TtlCacheStore(StoreABC[T]):
    store: StoreABC[T]
    cache: Dict[str, TtlEntry] = field(default_factory=dict)
    cached_result_sets: Dict[str, TtlEntry] = field(default_factory=dict)
    ttl: int = 30
    marshaller: MarshallerABC[T] = None

    def __post_init__(self):
        if not self.marshaller:
            object.__setattr__(
                self,
                "marshaller",
                get_default_context().get_marshaller(
                    self.get_meta().get_read_dataclass()
                ),
            )

    def clear_cache(self):
        self.cache.clear()
        self.cached_result_sets.clear()

    def get_meta(self) -> StoreMeta:
        return self.store.get_meta()

    def store_item_in_cache(self, key: str, item: T, expire_at: Optional[int] = None):
        if expire_at is None:
            expire_at = int(time()) + self.ttl
        self.cache[key] = TtlEntry(deepcopy(item), expire_at)

    def load_item_from_cache(self, key: str, now: Optional[int] = None):
        if now is None:
            now = int(time())
        entry = self.cache.get(key)
        if entry and entry.expire_at > now:
            return deepcopy(entry.value)

    def create(self, item: T) -> Optional[T]:
        item = self.store.create(item)
        key = self.get_meta().key_config.to_key_str(item)
        self.store_item_in_cache(key, item)
        return item

    def read(self, key: str) -> Optional[T]:
        item = self.load_item_from_cache(key)
        if item is None:
            item = self.store.read(key)
            if item:
                self.store_item_in_cache(key, item)
        return item

    def read_batch(self, keys: List[str]) -> List[Optional[T]]:
        now = int(time())
        items_by_key = {key: self.load_item_from_cache(key, now) for key in keys}
        keys_to_load = [key for key, item in items_by_key.items() if item is None]
        if keys_to_load:
            items = self.store.read_batch(keys_to_load)
            key_config = self.get_meta().key_config
            for item in items:
                if item:
                    key = key_config.to_key_str(item)
                    self.store_item_in_cache(key, item, now)
                    items_by_key[key] = item
        items = [items_by_key.get(key) for key in keys]
        return items

    # pylint: disable=W0212
    def _update(
        self,
        key: str,
        item: T,
        updates: T,
    ) -> Optional[T]:
        item = self.store._update(key, item, updates)
        if item:
            self.store_item_in_cache(key, item)
        return item

    # pylint: disable=W0212
    def _delete(self, key: str, item: T) -> bool:
        destroyed = self.store._delete(key, item)
        if key in self.cache:
            del self.cache[key]
        return destroyed

    def count(self, search_filter: SearchFilterABC = INCLUDE_ALL) -> int:
        return self.store.count(search_filter)

    def search(
        self,
        search_filter: SearchFilterABC = INCLUDE_ALL,
        search_order: Optional[SearchOrder] = None,
        page_key: Optional[str] = None,
        limit: Optional[int] = None,
    ) -> ResultSet[T]:
        result_set_key = [
            dump(search_filter, SearchFilterABC),
            dump(search_order, Optional[SearchOrder]),
            page_key,
            limit,
        ]
        result_set_key = secure_hash(result_set_key)
        now = int(time())
        entry = self.cached_result_sets.get(result_set_key)
        if entry and entry.expire_at > now:
            results = [
                self.load_item_from_cache(key, now) for key in entry.value["keys"]
            ]
            return ResultSet(results, entry.value["next_page_key"])
        result_set = self.store.search(search_filter, search_order, page_key, limit)
        key_config = self.get_meta().key_config
        expire_at = int(time()) + self.ttl
        keys = []
        for item in result_set.results:
            key = key_config.to_key_str(item)
            self.store_item_in_cache(key, item, expire_at)
            keys.append(key)
        entry = {"keys": keys, "next_page_key": result_set.next_page_key}
        self.cached_result_sets[result_set_key] = TtlEntry(entry, expire_at)
        return result_set

    def edit_batch(self, edits: List[BatchEdit]) -> List[BatchEditResult]:
        results = self.store.edit_batch(edits)
        for result in results:
            if not result.success:
                continue
            edit = result.edit
            if edit.update_item:
                key = self.get_meta().key_config.to_key_str(edit.update_item)
                self.cache.pop(key, None)
            elif edit.delete_key:
                self.cache.pop(edit.delete_key, None)
        return results



from dataclasses import dataclass
from typing import Dict, ForwardRef, Generic, List, Optional, TypeVar

from marshy.types import ExternalItemType
from schemey import schema_from_type
from servey.security.authorization import Authorization

from persisty.attr.attr import Attr, DEFAULT_PERMITTED_FILTER_OPS
from persisty.attr.attr_type import AttrType
from persisty.factory.store_factory_abc import StoreFactoryABC
from persisty.link.linked_store_abc import LinkedStoreABC
from persisty.link.on_delete import OnDelete

T = TypeVar("T")


class BelongsToCallable(Generic[T]):
    def __init__(self, key: str, store_factory: StoreFactoryABC):
        self.key = key
        self.store_factory = store_factory

    def __call__(self, authorization: Optional[Authorization] = None) -> Optional[T]:
        store = self.store_factory.create(authorization)
        item = store.read(self.key)
        return item


@dataclass
class BelongsTo(LinkedStoreABC, Generic[T]):
    name: Optional[str] = None
    key_attr_name: Optional[str] = None
    optional: Optional[bool] = None
    on_delete: OnDelete = OnDelete.BLOCK

    def get_name(self) -> str:
        return self.name

    def get_linked_type(self, forward_ref_ns: str) -> ForwardRef:
        return ForwardRef(
            forward_ref_ns + "." + self.get_linked_store_name().title().replace("_", "")
        )

    async def batch_call(
        self, keys: List, authorization: Optional[Authorization] = None
    ) -> List[Optional[T]]:
        if not keys:
            return []
        result = self.get_linked_store(authorization).read_batch(keys)
        return result

    def arg_extractor(self, obj):
        return [str(getattr(obj, self.key_attr_name))]

    def __set_name__(self, owner, name):
        self.name = name
        if self.key_attr_name is None:
            self.key_attr_name = f"{name}_id"

    def __get__(self, obj, obj_type) -> BelongsToCallable[T]:
        return BelongsToCallable(
            key=getattr(obj, self.key_attr_name),
            store_factory=self.get_linked_store_factory(),
        )

    def update_attrs(self, attrs_by_name: Dict[str, Attr]):
        if self.key_attr_name in attrs_by_name:
            return
        type_ = Optional[str] if self.optional else str
        schema = schema_from_type(type_)
        attrs_by_name[self.key_attr_name] = Attr(
            self.key_attr_name,
            AttrType.STR,
            schema,
            sortable=False,
            permitted_filter_ops=DEFAULT_PERMITTED_FILTER_OPS,
        )

    def update_json_schema(self, json_schema: ExternalItemType):
        id_attr_schema = json_schema.get("properties").get(self.key_attr_name)
        id_attr_schema["persistyBelongsTo"] = self.linked_store_name

        id_attr_schema = json_schema.get("properties").get(self.key_attr_name)
        id_attr_schema["persistyBelongsTo"] = self.linked_store_name



from servey.servey_celery import has_celery_broker

from persisty.store.store_abc import StoreABC
from persisty.trigger.store_triggers import StoreTriggers
from persisty.trigger.asyncio_trigger_store import AsyncioTriggerStore


def triggered_store(store: StoreABC):
    store_triggers = StoreTriggers(store.get_meta())
    if store_triggers.has_after_edit_actions():
        if has_celery_broker():
            from persisty.trigger.celery_trigger_store import CeleryTriggerStore

            store = CeleryTriggerStore(store, store_triggers)
        else:
            store = AsyncioTriggerStore(store, store_triggers)
    return store



from servey.util.singleton_abc import SingletonABC


class Undefined(SingletonABC):
    """
    Explicit value to be used for Undefined values. (Similar to Dataclasses MISSING but will not trigger an error on
    init of said dataclasses)

    For example, when a create / update is done but no value was specified, it just means we want these left alone, not
    that the update is Bad (Distinct from None, which can be a valid value for the stored)
    """

    def __bool__(self):
        return False  # Follow the javascript convention of making Undefined 'Falsy'

    def __eq__(self, other):
        return other is UNDEFINED

    def __repr__(self):
        return "undefined"

    def __hash__(self):
        return -1


UNDEFINED = Undefined()



from abc import ABC, abstractmethod
from typing import Optional, Tuple, Any

from sqlalchemy import Table

from persisty.search_filter.search_filter_abc import SearchFilterABC
from persisty.store_meta import StoreMeta


class SearchFilterConverterABC(ABC):
    priority: int = 100

    @abstractmethod
    def convert(
        self,
        search_filter: SearchFilterABC,
        table: Table,
        store_meta: StoreMeta,
        context,
    ) -> Optional[Tuple[Any, bool]]:
        """
        Convert to a sqlalchemy where clause. Return null if this is not possible. Returned boolean
        indicates whether the resulting clause fully handles this filter
        """



from __future__ import annotations
from abc import ABC, abstractmethod
from typing import Tuple, Optional, Any, TYPE_CHECKING, Generic, TypeVar

if TYPE_CHECKING:
    from persisty.attr.attr import Attr

T = TypeVar("T")


class SearchFilterABC(ABC, Generic[T]):
    @abstractmethod
    def lock_attrs(self, attrs: Tuple[Attr, ...]) -> SearchFilterABC:
        """Validate that this search_filter is applicable for the attrs given"""

    @abstractmethod
    def match(self, item: T, attrs: Tuple[Attr, ...]) -> bool:
        """Determine if the stored given matches this search_filter"""

    def __and__(self, obj_filter: SearchFilterABC) -> SearchFilterABC:
        if not isinstance(obj_filter, SearchFilterABC):
            raise TypeError(f"SearchFilterABC:{obj_filter}")
        from persisty.search_filter.and_filter import And

        return And((self, obj_filter))

    def __or__(self, obj_filter: SearchFilterABC) -> SearchFilterABC:
        if not isinstance(obj_filter, SearchFilterABC):
            raise TypeError(f"SearchFilterABC:{obj_filter}")
        from persisty.search_filter.or_filter import Or

        return Or((self, obj_filter))

    def __invert__(self) -> SearchFilterABC:
        from persisty.search_filter.not_filter import Not

        return Not(self)

    # pylint: disable=W0613
    def build_filter_expression(
        self, attrs: Tuple[Attr, ...]
    ) -> Tuple[Optional[Any], bool]:
        """
        Build a dynamodb filter expression from this search filter if possible, and return it. Return True if this
        filter was completely represented by the condition, False otherwise"""
        return None, False



from abc import ABC, abstractmethod
from typing import Optional, TypeVar, Generic, FrozenSet

from marshy.types import ExternalItemType

T = TypeVar("T")


class KeyConfigABC(ABC, Generic[T]):
    """Object for extracting / inserting string keys into stored items."""

    @abstractmethod
    def to_key_str(self, item: T) -> str:
        """Get the key from the stored given."""

    @abstractmethod
    def from_key_str(self, key: Optional[str], target: T):
        """Set the key for the stored given"""

    @abstractmethod
    def to_key_dict(self, key: Optional[str]) -> ExternalItemType:
        """Set the key for the stored given"""

    @abstractmethod
    def get_key_attrs(self) -> FrozenSet[str]:
        """Get the required attributes for this key"""



import importlib
import json
import os
from typing import Iterable, List
from importlib import resources

import marshy
from marshy.types import ExternalItemType

from persisty.store_meta import StoreMeta, T
from persisty.util import UNDEFINED


def get_seed_items(store_meta: StoreMeta) -> Iterable[T]:
    # noinspection PyBroadException
    try:
        seed_data = get_seed_data(store_meta.name)
        item_dataclass = store_meta.get_stored_dataclass()
        seed_items = []
        for data in seed_data:
            item = marshy.load(item_dataclass, data)
            for attr in store_meta.attrs:
                if attr.create_generator and getattr(item, attr.name) is UNDEFINED:
                    setattr(
                        item,
                        attr.name,
                        attr.create_generator.transform(UNDEFINED, item),
                    )
            seed_items.append(item)
        return seed_items
    except Exception as e:
        print(e)
        return []


def get_seed_data(name: str) -> List[ExternalItemType]:
    seeds_package = os.environ.get("SEEDS_PACKAGE") or "seeds"
    seeds = importlib.import_module(seeds_package)
    items = getattr(seeds, name, None)
    if items is not None:
        return items
    seed_data = resources.read_text(seeds_package, name + ".json")
    seed_data = json.loads(seed_data)
    return seed_data



import os
from dataclasses import dataclass, field
from typing import Optional, Dict, Set, Iterator, List

import boto3
from schemey import schema_from_type

from persisty.attr.attr import Attr
from persisty.attr.attr_filter_op import TYPE_FILTER_OPS
from persisty.attr.attr_type import attr_type, AttrType
from persisty.errors import PersistyError
from persisty.impl.dynamodb.partition_sort_index import PartitionSortIndex, from_schema
from persisty.impl.dynamodb.dynamodb_table_store import DynamodbTableStore
from persisty.index.attr_index import AttrIndex
from persisty.key_config.attr_key_config import AttrKeyConfig
from persisty.key_config.composite_key_config import CompositeKeyConfig
from persisty.key_config.key_config_abc import KeyConfigABC
from persisty.store.restrict_access_store import restrict_access_store
from persisty.store.schema_validating_store import SchemaValidatingStore
from persisty.store.store_abc import StoreABC
from persisty.store.unique_index_store import unique_index_store
from persisty.store_meta import StoreMeta
from persisty.util import filter_none


@dataclass
class DynamodbStoreFactory:
    meta: Optional[StoreMeta] = None
    aws_profile_name: Optional[str] = None
    region_name: Optional[str] = field(
        default_factory=lambda: os.environ.get("AWS_REGION")
    )
    table_name: Optional[str] = None
    index: Optional[PartitionSortIndex] = None
    global_secondary_indexes: Optional[Dict[str, PartitionSortIndex]] = None

    def get_meta(self) -> StoreMeta:
        return self.meta

    def create(self) -> Optional[StoreABC]:
        store = DynamodbTableStore(
            meta=self.meta,
            table_name=self.table_name,
            index=self.index,
            global_secondary_indexes=self.global_secondary_indexes or {},
            aws_profile_name=self.aws_profile_name,
            region_name=self.region_name,
        )
        store = SchemaValidatingStore(store)
        store = restrict_access_store(store, self.meta.store_access)
        store = unique_index_store(store)
        return store

    def derive_from_meta(self):
        meta = self.meta
        if self.table_name is None:
            self.table_name = meta.name
        if self.index is None:
            key_config = meta.key_config
            key_config_attrs = list(_get_attrs_from_key(key_config))
            self.index = PartitionSortIndex(*key_config_attrs)
        if self.global_secondary_indexes is None:
            self.global_secondary_indexes = {}
            for index in meta.indexes:
                if isinstance(index, AttrIndex):
                    self.global_secondary_indexes[
                        f"gix__{index.attr_name}"
                    ] = PartitionSortIndex(index.attr_name)
                elif isinstance(index, PartitionSortIndex):
                    if index.sk:
                        self.global_secondary_indexes[
                            f"gix__{index.pk}__{index.sk}"
                        ] = index
                    else:
                        self.global_secondary_indexes[f"gix__{index.pk}"] = index

    def get_session(self):
        kwargs = filter_none(
            {"profile_name": self.aws_profile_name, "region_name": self.region_name}
        )
        session = boto3.Session(**kwargs)
        return session

    def load_from_aws(self):
        dynamodb = self.get_session().client("dynamodb")
        table_meta = dynamodb.describe_table(TableName=self.table_name)
        table = table_meta["Table"]
        self.index = from_schema(table["KeySchema"])
        self.global_secondary_indexes = {
            i["IndexName"]: from_schema(i["KeySchema"])
            for i in (table.get("GlobalSecondaryIndexes") or [])
            if i["Projection"]["ProjectionType"] == "ALL"
            and i["IndexStatus"] == "ACTIVE"
        }
        attrs = tuple(
            _dynamo_attr_to_attr(a) for a in (table.get("AttributeDefinitions") or [])
        )
        key_config = self.index.key_config_from_attrs(attrs)
        self.meta = StoreMeta(
            name=self.table_name,
            attrs=attrs,
            key_config=key_config,
        )

    def create_table_in_aws(self):
        dynamodb = self.get_session().client("dynamodb")
        kwargs = {
            "AttributeDefinitions": self.get_attribute_definitions(),
            "TableName": self.table_name,
            "KeySchema": self.index.to_schema(),
            "BillingMode": "PAY_PER_REQUEST",  # Ops teams will want to look at these values
        }
        if self.global_secondary_indexes:
            kwargs["GlobalSecondaryIndexes"] = self.get_global_secondary_indexes()
        response = dynamodb.create_table(**kwargs)
        return response

    def get_attribute_definitions(self) -> List[Dict]:
        attrs = {}
        self._attrs(self.index, attrs)
        if self.global_secondary_indexes:
            for index in self.global_secondary_indexes.values():
                self._attrs(index, attrs)
        return list(attrs.values())

    def get_global_secondary_indexes(self):
        return [
            {
                "IndexName": k,
                "KeySchema": i.to_schema(),
                "Projection": {"ProjectionType": "ALL"},
            }
            for k, i in (self.global_secondary_indexes or {}).items()
        ]

    def _attrs(self, index: PartitionSortIndex, attrs: Dict):
        attrs[index.pk] = self._attr(index.pk)
        if index.sk:
            attrs[index.sk] = self._attr(index.sk)

    def _attr(self, name: str):
        attr = next(a for a in self.meta.attrs if a.name == name)
        return {
            "AttributeName": name,
            "AttributeType": _FIELD_TYPE_2_DYNAMODB[attr.attr_type],
        }


def _remove_index(indexed_attrs: Set[str], index: PartitionSortIndex):
    indexed_attrs.remove(index.pk)
    if index.sk:
        indexed_attrs.remove(index.sk)


_DYNAMODB_2_PYTHON = {
    "B": bytearray,
    "N": float,
    "S": str,
}

_FIELD_TYPE_2_DYNAMODB = {
    AttrType.BINARY: "B",
    AttrType.STR: "S",
    AttrType.FLOAT: "N",
    AttrType.INT: "N",
    AttrType.DATETIME: "S",
    AttrType.UUID: "S",
}


def _dynamo_attr_to_attr(dynamo_attr: Dict):
    python_type = _DYNAMODB_2_PYTHON[dynamo_attr["AttributeType"]]
    db_type = attr_type(python_type)
    attr = Attr(
        name=dynamo_attr["AttributeName"],
        attr_type=db_type,
        schema=schema_from_type(python_type),
        permitted_filter_ops=TYPE_FILTER_OPS.get(db_type),
    )
    return attr


def _get_attrs_from_key(key_config: KeyConfigABC) -> Iterator[str]:
    if isinstance(key_config, AttrKeyConfig):
        yield key_config.attr_name
    elif isinstance(key_config, CompositeKeyConfig):
        for k in key_config.attrs:
            yield from _get_attrs_from_key(k)
    else:
        raise PersistyError("unsupported_key")



from typing import Optional, Tuple, Any

from sqlalchemy import Table, not_

from persisty.impl.sqlalchemy.search_filter.and_filter_converter import (
    AndFilterConverter,
)
from persisty.search_filter.not_filter import Not
from persisty.search_filter.search_filter_abc import SearchFilterABC
from persisty.store_meta import StoreMeta


class NotFilterConverter(AndFilterConverter):
    def convert(
        self,
        search_filter: SearchFilterABC,
        table: Table,
        store_meta: StoreMeta,
        context,
    ) -> Optional[Tuple[Any, bool]]:
        if isinstance(search_filter, Not):
            clause, handled = context.convert(
                search_filter.search_filter, table, store_meta
            )
            if clause:
                clause = not_(clause)
            return clause, handled



from __future__ import annotations
from dataclasses import dataclass
from typing import Optional, Type, Generic

from persisty.batch_edit import BatchEdit, C, U


@dataclass
class BatchEditResult(Generic[C, U]):
    edit: BatchEdit[C, U]
    success: bool = False
    code: Optional[str] = None
    details: Optional[str] = None

    def copy_from(self, result: BatchEditResult):
        self.edit = result.edit
        self.success = result.success
        self.code = result.code
        self.details = result.details


def batch_edit_result_dataclass_for(batch_edit_type: Type) -> Type:
    params = {
        "__annotations__": {
            "edit": batch_edit_type,
            "success": bool,
            "code": Optional[str],
            "details": Optional[str],
        },
        "__doc__": f"Batch edit result of {batch_edit_type.__name__}",
        "success": False,
        "code": None,
        "details": None,
    }
    type_name = f"{batch_edit_type.__name__}Result"
    type_ = dataclass(type(type_name, (), params))
    return type_



from dataclasses import dataclass, Field, field
from enum import Enum
from typing import Optional, Tuple

from schemey import Schema, schema_from_type

from persisty.attr.attr_filter_op import AttrFilterOp, TYPE_FILTER_OPS
from persisty.attr.attr_type import AttrType, attr_type, ATTR_TYPE_MAP
from persisty.attr.generator.attr_value_generator_abc import AttrValueGeneratorABC

from persisty.util.undefined import UNDEFINED


# pylint: disable=R0902
@dataclass
class Attr:
    name: str = UNDEFINED  # Can be populated by __set_name__
    attr_type: AttrType = UNDEFINED
    schema: Schema = UNDEFINED
    creatable: bool = True
    readable: bool = True
    updatable: bool = True
    sortable: bool = True
    create_generator: Optional[AttrValueGeneratorABC] = None
    update_generator: Optional[AttrValueGeneratorABC] = None
    permitted_filter_ops: Tuple[AttrFilterOp, ...] = UNDEFINED

    def __set_name__(self, owner, name):
        if self.name is UNDEFINED:
            self.name = name
        annotations = owner.__dict__.get("__annotations__")
        type_ = annotations.get(name)
        if self.attr_type is UNDEFINED:
            self.attr_type = attr_type(type_)
        if self.schema is UNDEFINED:
            self.schema = schema_from_type(type_)
        if self.permitted_filter_ops is UNDEFINED:
            self.permitted_filter_ops = (
                TYPE_FILTER_OPS.get(self.attr_type) or DEFAULT_PERMITTED_FILTER_OPS
            )

    def to_field(self) -> Field:
        result = field(
            default=UNDEFINED, metadata={"schemey": self.schema, "persisty": self}
        )
        result.name = self.name
        result.type = self.schema.python_type
        return result

    def sanitize_type(self, value):
        """
        Sanitize the type of an attribute value - try and make sure the type is correct.
        """
        if value in (None, UNDEFINED):
            return value
        type_ = ATTR_TYPE_MAP[self.attr_type]
        if isinstance(value, type_):
            return value
        if isinstance(value, Enum) and isinstance(value.value, type_):
            return value
        value = type_(value)
        return value


DEFAULT_PERMITTED_FILTER_OPS = (
    AttrFilterOp.eq,
    AttrFilterOp.exists,
    AttrFilterOp.ne,
    AttrFilterOp.not_exists,
)



import inspect
from dataclasses import dataclass
from enum import Enum
from typing import Optional, Iterator

from marshy.factory.optional_marshaller_factory import get_optional_type
from marshy.types import ExternalItemType
from sqlalchemy import (
    Column,
    Integer,
    Float,
    String,
    Text,
    DateTime,
    Boolean,
    Enum as SqlalchemyEnum,
    LargeBinary,
)

# noinspection PyPep8Naming
from sqlalchemy.dialects.postgresql import UUID as PostgresUuid, JSON as PostgresJson

# noinspection PyPep8Naming
from sqlalchemy.dialects.mysql import JSON as MysqlJson

# noinspection PyPep8Naming
from sqlalchemy.dialects.mssql import JSON as MssqlJson

from persisty.attr.attr import Attr
from persisty.key_config.key_config_abc import KeyConfigABC

POSTGRES = "postgres"
MYSQL = "mysql"
MSSQL = "mssql"


@dataclass
class SqlalchemyColumnConverter:
    key_config: KeyConfigABC
    dialect: str

    def create_column(self, attr: Attr) -> Column:
        fn_name = f"_create_{attr.attr_type.value}"
        args = [attr.name, getattr(self, fn_name)(attr)]
        kwargs = {
            "nullable": bool(get_optional_type(attr.schema.python_type)),
        }
        if attr.name in self.key_config.get_key_attrs():
            kwargs["primary_key"] = True
        column = Column(*args, **kwargs)
        return column

    # pylint: disable=W0613
    # noinspection PyUnusedLocal
    @staticmethod
    def _create_binary(attr: Attr):
        return LargeBinary

    # pylint: disable=W0613
    # noinspection PyUnusedLocal
    @staticmethod
    def _create_bool(attr: Attr):
        return Boolean

    # pylint: disable=W0613
    # noinspection PyUnusedLocal
    @staticmethod
    def _create_datetime(attr: Attr):
        return DateTime

    # pylint: disable=W0613
    # noinspection PyUnusedLocal
    @staticmethod
    def _create_float(attr: Attr):
        return Float

    # pylint: disable=W0613
    # noinspection PyUnusedLocal
    @staticmethod
    def _create_int(attr: Attr):
        return Integer

    # pylint: disable=W0613
    # noinspection PyUnusedLocal
    def _create_json(self, attr: Attr):
        if self.dialect == POSTGRES:
            return PostgresJson
        if self.dialect == MYSQL:
            return MysqlJson
        if self.dialect == MSSQL:
            return MssqlJson
        return Text

    def _create_str(self, attr: Attr):
        type_ = attr.schema.python_type
        type_ = get_optional_type(type_) or type_
        if inspect.isclass(type_) and issubclass(type_, Enum):
            return SqlalchemyEnum(type_)
        length = self.get_column_length_from_schema(attr.schema.schema)
        if not length:
            return Text
        return String(length=length)

    # pylint: disable=W0613
    # noinspection PyUnusedLocal
    def _create_uuid(self, attr: Attr):
        # AFAIK at the moment, the only dialect with a native UUID type is postgres.
        # So the others will define UUIDs as strings (for readability)
        if self.dialect == POSTGRES:
            return PostgresUuid(as_uuid=True)
        return String(length=36)

    def get_column_length_from_schema(
        self, json_schema: ExternalItemType
    ) -> Optional[int]:
        try:
            return max(self.get_max_lengths_from_schema(json_schema))
        except ValueError:
            return None

    def get_max_lengths_from_schema(self, json_schema) -> Iterator[int]:
        if isinstance(json_schema, dict):
            max_length = json_schema.get("maxLength")
            if max_length:
                yield max_length
            for value in json_schema.values():
                yield from self.get_max_lengths_from_schema(value)
        if isinstance(json_schema, list):
            for value in json_schema:
                yield from self.get_max_lengths_from_schema(value)



from typing import Type
from dataclasses import dataclass

from persisty.attr.attr import Attr
from persisty.attr.attr_filter import AttrFilter
from persisty.attr.attr_filter_op import AttrFilterOp
from persisty.errors import PersistyError
from persisty.search_order.search_order import SearchOrder

from persisty.search_order.search_order_attr import SearchOrderAttr
from persisty.store_meta import get_meta, StoreMeta


@dataclass
class ObjFilterFactory:
    meta: StoreMeta

    def __getattr__(self, name):
        try:
            attr = next(a for a in self.meta.attrs if a.name == name and a.readable)
            return AttrFilterFactory(attr)
        except StopIteration as exc:
            raise PersistyError(f"no_such_field:{name}") from exc


@dataclass
class AttrFilterFactory:
    attr: Attr

    def asc(self) -> SearchOrder:
        return SearchOrder((SearchOrderAttr(self.attr.name),))

    def desc(self):
        return SearchOrder((SearchOrderAttr(self.attr.name, True),))

    def __getattr__(self, name):
        op = next(op for op in self.attr.permitted_filter_ops if op.name == name)
        return AttrOpFilterFactory(self.attr, op)


@dataclass
class AttrOpFilterFactory:
    attr: Attr
    op: AttrFilterOp

    def __call__(self, value):
        return AttrFilter(self.attr.name, self.op, value)


def filter_factory(type_: Type) -> ObjFilterFactory:
    return ObjFilterFactory(get_meta(type_))



from typing import Dict

from persisty.factory.store_factory_abc import StoreFactoryABC
from persisty.servey import generated


def add_actions_for_all_store_factories(target: Dict):
    from persisty.finder.store_finder_abc import find_store_factories

    for store_factory in find_store_factories():
        add_actions_for_store_factory(store_factory, target)


def add_actions_for_store_factory(store_factory: StoreFactoryABC, target: Dict):
    for action_ in store_factory.create_actions():
        target[action_.name] = action_.fn



from abc import abstractmethod, ABC
from typing import Iterator

from marshy.factory.impl_marshaller_factory import get_impls

from persisty.factory.store_factory_abc import StoreFactoryABC
from persisty.store.store_abc import StoreABC


class StoreFactoryFinderABC(ABC):
    @abstractmethod
    def find_stores(self) -> Iterator[StoreABC]:
        """Find all available store items"""

    @abstractmethod
    def find_store_factories(self) -> Iterator[StoreFactoryABC]:
        """Find all available store items"""


def find_stores() -> Iterator[StoreABC]:
    names = set()
    for finder in get_impls(StoreFactoryFinderABC):
        for store in finder().find_stores():
            name = store.get_meta().name
            if name not in names:
                names.add(name)
                yield store


def find_store_factories() -> Iterator[StoreFactoryABC]:
    names = set()
    for finder in get_impls(StoreFactoryFinderABC):
        for factory in finder().find_store_factories():
            name = factory.get_meta().name
            if name not in names:
                names.add(name)
                yield factory



from abc import ABC, abstractmethod
from itertools import islice
from typing import Optional, List, Iterator, Dict, Generic

from persisty.errors import PersistyError
from persisty.batch_edit import BatchEdit
from persisty.batch_edit_result import BatchEditResult
from persisty.result_set import ResultSet
from persisty.search_filter.include_all import INCLUDE_ALL
from persisty.search_filter.search_filter_abc import SearchFilterABC
from persisty.search_order.search_order import SearchOrder
from persisty.store_meta import StoreMeta, T


class StoreABC(Generic[T], ABC):
    """
    General contract for storage object, allowing CRUD, search, and batch updates for objects
    """

    @abstractmethod
    def get_meta(self) -> StoreMeta:
        """Get the meta for this storage"""

    @abstractmethod
    def create(self, item: T) -> Optional[T]:
        """Create an stored in the data store"""

    @abstractmethod
    def read(self, key: str) -> Optional[T]:
        """Read an stored from the data store"""

    def read_batch(self, keys: List[str]) -> List[Optional[T]]:
        assert len(keys) <= self.get_meta().batch_size
        items = [self.read(key) for key in keys]
        return items

    def read_all(self, keys: Iterator[str]) -> Iterator[Optional[T]]:
        keys = iter(keys)
        while True:
            batch_keys = list(islice(keys, self.get_meta().batch_size))
            if not batch_keys:
                return
            items = self.read_batch(batch_keys)
            yield from items

    def update(
        self, updates: T, precondition: SearchFilterABC = INCLUDE_ALL
    ) -> Optional[T]:
        """
        Update (a partial set of values from) an item based upon its key and the constraint given. By convention
        any UNDEFINED value is ignored. Return the full new version of the item if an update occured. If the key
        extracted from the updates did not match any existing item, return None. If any other error occurrecd, throw
        a PersistyError
        """
        key = self.get_meta().key_config.to_key_str(updates)
        if not key:
            raise PersistyError(f"missing_key:{updates}")
        item = self.read(key)
        attrs = self.get_meta().attrs
        precondition = precondition.lock_attrs(attrs)
        if item and precondition.match(item, attrs):
            return self._update(key, item, updates)

    @abstractmethod
    def _update(
        self,
        key: str,
        item: T,
        updates: T,
    ) -> Optional[T]:
        """
        Update (a partial set of values from) an item based upon its key and the constraint given. By convention
        any UNDEFINED value is ignored. Return the full new version of the item if an update occured. If the key
        extracted from the updates did not match any existing item, return None. If any other error occurrecd, throw
        a PersistyError
        """

    def delete(self, key: str) -> bool:
        """Delete an stored from the data store. Return true if an item was deleted, false otherwise"""
        key = str(key)
        item = self.read(key)
        if not item:
            return False
        return self._delete(key, item)

    @abstractmethod
    def _delete(self, key: str, item: T) -> bool:
        """Delete an stored from the data store. Return true if an item was deleted, false otherwise"""

    def search(
        self,
        search_filter: SearchFilterABC[T] = INCLUDE_ALL,
        search_order: Optional[SearchOrder[T]] = None,
        page_key: Optional[str] = None,
        limit: Optional[int] = None,
    ) -> ResultSet[T]:
        if limit is None:
            limit = self.get_meta().batch_size
        assert limit <= self.get_meta().batch_size
        items = self.search_all(search_filter, search_order)
        skip_to_page(page_key, items, self.get_meta().key_config)
        items = list(islice(items, limit))
        page_key = None
        if len(items) == limit:
            page_key = self.get_meta().key_config.to_key_str(items[-1])
        return ResultSet(items, page_key)

    def search_all(
        self,
        search_filter: SearchFilterABC[T] = INCLUDE_ALL,
        search_order: Optional[SearchOrder[T]] = None,
    ) -> Iterator[T]:
        page_key = None
        while True:
            result_set = self.search(search_filter, search_order, page_key)
            yield from result_set.results
            page_key = result_set.next_page_key
            if not page_key:
                return

    @abstractmethod
    def count(self, search_filter: SearchFilterABC[T] = INCLUDE_ALL) -> int:
        """Create an stored in the data store"""

    def edit_batch(self, edits: List[BatchEdit[T, T]]) -> List[BatchEditResult[T, T]]:
        """
        Do a batch edit and return a list of results. The results should contain all the same edits in the same
        order
        """
        assert len(edits) <= self.get_meta().batch_size
        to_key_str = self.get_meta().key_config.to_key_str
        keys = []
        for edit in edits:
            if edit.create_item:
                key = to_key_str(edit.create_item)
                if key:
                    keys.append(key)
            if edit.update_item:
                keys.append(to_key_str(edit.update_item))
            elif edit.delete_key:
                keys.append(edit.delete_key)
        items_by_key = {
            to_key_str(item): item for item in self.read_batch(keys) if item
        }
        filtered_edits = []
        for edit in edits:
            if edit.create_item:
                key = to_key_str(edit.create_item)
                if key:
                    item = items_by_key.get(key)
                    if item:
                        continue
                filtered_edits.append(edit)
                continue
            if edit.update_item:
                key = to_key_str(edit.update_item)
                if key in items_by_key:
                    filtered_edits.append(edit)
            elif edit.delete_key and edit.delete_key in items_by_key:
                filtered_edits.append(edit)
        filtered_results = self._edit_batch(filtered_edits, items_by_key)
        filtered_results_by_id = {r.edit.id: r for r in filtered_results}
        results = [
            filtered_results_by_id.get(e.id)
            or BatchEditResult(
                e, False, "duplicate_key" if e.create_item else "missing_key"
            )
            for e in edits
        ]
        return results

    def _edit_batch(
        self, edits: List[BatchEdit[T, T]], items_by_key: Dict[str, T]
    ) -> List[BatchEditResult[T, T]]:
        """
        Simple non transactional implementation of batch functionality. Other implementations employ strategies to boost
        performance such reducing the number of network round trips. Whether an edit is atomic is dependant on the
        underlying mechanism, but should be reflected in the results.
        """
        results = []
        to_key_str = self.get_meta().key_config.to_key_str
        for edit in edits:
            try:
                if edit.create_item:
                    item = self.create(edit.create_item)
                    results.append(BatchEditResult(edit, bool(item)))
                elif edit.update_item:
                    key = to_key_str(edit.update_item)
                    item = items_by_key[key]
                    item = self._update(key, item, edit.update_item)
                    results.append(BatchEditResult(edit, bool(item)))
                else:
                    item = items_by_key.get(edit.delete_key)
                    deleted = self._delete(edit.delete_key, item)
                    results.append(BatchEditResult(edit, bool(deleted)))
            except Exception as e:
                results.append(BatchEditResult(edit, False, "exception", str(e)))
        return results

    def edit_all(
        self, edits: Iterator[BatchEdit[T, T]]
    ) -> Iterator[BatchEditResult[T, T]]:
        edits = iter(edits)
        while True:
            page = list(islice(edits, self.get_meta().batch_size))
            if not page:
                break
            results = self.edit_batch(page)
            yield from results


def skip_to_page(page_key: str, items, key_config):
    if page_key:
        while True:
            next_result = next(items, None)
            if next_result is None:
                raise PersistyError("invalid_page_key")  # The item was probably deleted
            key = key_config.to_key_str(next_result)
            if key == page_key:
                return



from logging import Logger
from time import time
from typing import Callable, Optional, TypeVar, FrozenSet, Any

from persisty.util import get_logger, filter_none

T = TypeVar("T")


def logify(
    obj: T,
    log_methods: Optional[FrozenSet[str]] = None,
    logger: Optional[Logger] = None,
) -> T:
    return LoggingWrapper(obj, log_methods, logger)


def logify_callable(name, fn: Callable, logger: Optional[Logger] = None):
    def wrapper(*args, **kwargs):
        start = time()
        error = None
        try:
            return_value = fn(*args, **kwargs)
            return return_value
        except Exception as e:
            error = str(e)
            raise e
        finally:
            end = time()
            time_taken = round(end - start)
            msg = filter_none(
                {
                    "name": name,
                    "time": time_taken,
                    "error": error,
                    "args": args if args else None,
                    "kwargs": kwargs if kwargs else None,
                }
            )
            logger.info(msg)

    return wrapper


class LoggingWrapper:
    def __init__(
        self,
        wrapped: Any,
        log_methods: Optional[FrozenSet[str]] = None,
        logger: Optional[Logger] = None,
    ):
        self.wrapped = wrapped
        if logger is None:
            logger = get_logger(wrapped.__class__.__name__)
        self.logger = logger
        self.log_methods = log_methods

    def __getattr__(self, name: str):
        ret = getattr(self.wrapped, name)
        if callable(ret):
            if self.log_methods:
                wrap = name in self.log_methods
            else:
                wrap = not name.startswith("_")
            if wrap:
                ret = logify_callable(name, ret, self.logger)
                setattr(self, name, ret)
        return ret



import dataclasses
from typing import Optional, Iterator, List, Dict

from servey.security.authorization import Authorization

from persisty.attr.attr_filter import AttrFilter
from persisty.attr.attr_filter_op import AttrFilterOp
from persisty.batch_edit import BatchEdit
from persisty.batch_edit_result import BatchEditResult
from persisty.result_set import ResultSet
from persisty.search_filter.include_all import INCLUDE_ALL
from persisty.search_filter.search_filter_abc import SearchFilterABC
from persisty.search_order.search_order import SearchOrder
from persisty.store.store_abc import StoreABC, T
from persisty.store_meta import StoreMeta


@dataclasses.dataclass
class OwnedStore(StoreABC[T]):
    """
    Store which enforces ownership of items within it. Each item has an attribute
    which maps to the id of the owner, and access to read, update and delete operations
    can be optionally restricted
    """

    store: StoreABC[T]
    authorization: Authorization
    subject_id_attr_name: str = "subject_id"
    require_ownership_for_read: bool = False
    require_ownership_for_update: bool = True
    require_ownership_for_delete: bool = True

    def get_meta(self) -> StoreMeta:
        meta = getattr(self, "_meta", None)
        if meta is None:
            meta = meta_with_non_editable_subject_id(
                self.store.get_meta(), self.subject_id_attr_name
            )
            setattr(self, "_meta", meta)
        return meta

    def create(self, item: T) -> Optional[T]:
        setattr(item, self.subject_id_attr_name, self.authorization.subject_id)
        return self.store.create(item)

    def read(self, key: str) -> Optional[T]:
        item = self.store.read(key)
        if item and self.require_ownership_for_read:
            if (
                getattr(item, self.subject_id_attr_name)
                != self.authorization.subject_id
            ):
                return
        return item

    # pylint: disable=W0212
    def _update(self, key: str, item: T, updates: T) -> Optional[T]:
        if self.require_ownership_for_update:
            if (
                getattr(item, self.subject_id_attr_name)
                != self.authorization.subject_id
            ):
                return
        setattr(updates, self.subject_id_attr_name, self.authorization.subject_id)
        return self.store._update(key, item, updates)

    # pylint: disable=W0212
    def _delete(self, key: str, item: T) -> bool:
        if self.require_ownership_for_delete:
            if (
                getattr(item, self.subject_id_attr_name)
                != self.authorization.subject_id
            ):
                return False
        return self.store._delete(key, item)

    def count(self, search_filter: SearchFilterABC[T] = INCLUDE_ALL) -> int:
        if self.require_ownership_for_read:
            search_filter &= AttrFilter(
                self.subject_id_attr_name,
                AttrFilterOp.eq,
                self.authorization.subject_id,
            )
        return self.store.count(search_filter)

    def search(
        self,
        search_filter: SearchFilterABC[T] = INCLUDE_ALL,
        search_order: Optional[SearchOrder[T]] = None,
        page_key: Optional[str] = None,
        limit: Optional[int] = None,
    ) -> ResultSet[T]:
        if self.require_ownership_for_read:
            search_filter &= AttrFilter(
                self.subject_id_attr_name,
                AttrFilterOp.eq,
                self.authorization.subject_id,
            )
        return self.store.search(search_filter, search_order, page_key, limit)

    def search_all(
        self,
        search_filter: SearchFilterABC[T] = INCLUDE_ALL,
        search_order: Optional[SearchOrder[T]] = None,
    ) -> Iterator[T]:
        if self.require_ownership_for_read:
            search_filter &= AttrFilter(
                self.subject_id_attr_name,
                AttrFilterOp.eq,
                self.authorization.subject_id,
            )
        return self.store.search_all(search_filter, search_order)

    def _edit_batch(
        self, edits: List[BatchEdit[T, T]], items_by_key: Dict[str, T]
    ) -> List[BatchEditResult[T, T]]:
        filtered_edits = []
        to_key_str = self.get_meta().key_config.to_key_str
        for edit in edits:
            if edit.create_item:
                setattr(
                    edit.create_item,
                    self.subject_id_attr_name,
                    self.authorization.subject_id,
                )
                filtered_edits.append(edit)
            elif edit.update_item:
                key = to_key_str(edit.update_item)
                existing_item = items_by_key.get(key)
                existing_subject_id = getattr(existing_item, self.subject_id_attr_name)
                if (
                    self.require_ownership_for_update
                    and existing_subject_id != self.authorization.subject_id
                ):
                    continue
                setattr(
                    edit.create_item,
                    self.subject_id_attr_name,
                    self.authorization.subject_id,
                )
                filtered_edits.append(edit)
            else:
                existing_item = items_by_key.get(edit.delete_key)
                existing_subject_id = getattr(existing_item, self.subject_id_attr_name)
                if (
                    self.require_ownership_for_delete
                    and existing_subject_id != self.authorization.subject_id
                ):
                    continue
                filtered_edits.append(edit)
        filtered_results = self.store._edit_batch(filtered_edits, items_by_key)
        results_by_id = {r.edit.id: r for r in filtered_results}
        results = []
        for edit in edits:
            result = results_by_id.get(edit.id)
            if result:
                results.append(result)
            else:
                results.append(BatchEditResult(edit, False, "exception", "missing_key"))
        return results


def meta_with_non_editable_subject_id(
    meta: StoreMeta, subject_id_attr_name: str
) -> StoreMeta:
    attrs = []
    for attr in meta.attrs:
        if attr.name == subject_id_attr_name:
            attr = dataclasses.replace(attr, creatable=False, updatable=False)
        attrs.append(attr)
    return dataclasses.replace(meta, attrs=attrs)



from dataclasses import dataclass
from typing import Optional, Generic, TypeVar, ForwardRef

from servey.security.authorization import Authorization

from persisty.attr.attr_filter import AttrFilter, AttrFilterOp
from persisty.factory.store_factory_abc import StoreFactoryABC
from persisty.link.linked_store_abc import LinkedStoreABC
from persisty.result_set import ResultSet
from persisty.search_filter.search_filter_abc import SearchFilterABC
from persisty.search_order.search_order import SearchOrder
from persisty.util import to_snake_case

T = TypeVar("T")


@dataclass
class HasManyCallable(Generic[T]):
    store_factory: StoreFactoryABC
    search_filter: SearchFilterABC
    search_order: Optional[SearchOrder] = None
    limit: Optional[int] = None

    def __call__(self, authorization: Optional[Authorization] = None) -> ResultSet[T]:
        store = self.store_factory.create(authorization)
        result_set = store.search(
            search_filter=self.search_filter,
            search_order=self.search_order,
            limit=self.limit,
        )
        return result_set


@dataclass
class HasMany(LinkedStoreABC, Generic[T]):
    name: Optional[str] = None  # Allows None so __set_name__ can exist
    key_attr_name: Optional[str] = None
    local_key_attr_name: str = "id"
    remote_key_attr_name: Optional[str] = None
    limit: int = 10
    search_order: Optional[SearchOrder] = None

    def __set_name__(self, owner, name):
        self.name = name
        if self.remote_key_attr_name is None:
            self.remote_key_attr_name = f"{to_snake_case(owner.__name__)}_id"

    def get_name(self) -> str:
        return self.name

    def get_linked_type(self, forward_ref_ns: str) -> ForwardRef:
        return ForwardRef(
            forward_ref_ns
            + "."
            + self.get_linked_store_name().title().replace("_", "")
            + "ResultSet"
        )

    def __get__(self, obj, obj_type) -> HasManyCallable[T]:
        key = getattr(obj, self.local_key_attr_name)
        return HasManyCallable(
            store_factory=self.get_linked_store_factory(),
            search_filter=AttrFilter(self.remote_key_attr_name, AttrFilterOp.eq, key),
            search_order=self.search_order,
        )



from dataclasses import dataclass, field
from typing import List, Iterator, Awaitable

from marshy import get_default_context
from marshy.marshaller_context import MarshallerContext
from servey.action.action import Action
from servey.finder.action_finder_abc import find_actions_with_trigger_type

from persisty.store_meta import StoreMeta
from persisty.trigger.after_create_trigger import AfterCreateTrigger
from persisty.trigger.after_delete_trigger import AfterDeleteTrigger
from persisty.trigger.after_update_trigger import AfterUpdateTrigger


@dataclass
class StoreTriggers:
    """
    Object for collecting / coordingating triggers to be run after edits to a store object.
    The actual mechanism for running triggers may be asyncio, celery, or even a post processing
    lambda on a dynamodb table.
    """

    store_meta: StoreMeta
    after_create_actions: List[Action] = None
    after_update_actions: List[Action] = None
    after_delete_actions: List[Action] = None
    marshaller_context: MarshallerContext = field(default_factory=get_default_context)

    def get_after_create_actions(self):
        if self.after_create_actions is None:
            self.after_create_actions = list(
                _get_triggered_actions(self.store_meta.name, AfterCreateTrigger)
            )
        return self.after_create_actions

    def get_after_update_actions(self):
        if self.after_create_actions is None:
            self.after_create_actions = list(
                _get_triggered_actions(self.store_meta.name, AfterUpdateTrigger)
            )
        return self.after_create_actions

    def get_after_delete_actions(self):
        if self.after_create_actions is None:
            self.after_create_actions = list(
                _get_triggered_actions(self.store_meta.name, AfterDeleteTrigger)
            )
        return self.after_create_actions

    def has_after_create_actions(self):
        return bool(self.get_after_create_actions())

    def has_after_update_actions(self):
        return bool(self.get_after_update_actions())

    def has_after_delete_actions(self):
        return bool(self.get_after_delete_actions())

    def has_after_edit_actions(self):
        return (
            self.has_after_create_actions()
            or self.has_after_update_actions()
            or self.has_after_delete_actions()
        )

    async def async_after_create(self, new_item):
        for action in self.get_after_create_actions():
            result = action.fn(new_item)
            if isinstance(result, Awaitable):
                await result

    async def async_after_update(self, old_item, new_item):
        for action in self.get_after_create_actions():
            result = action.fn(old_item, new_item)
            if isinstance(result, Awaitable):
                await result

    async def async_after_delete(self, old_item):
        for action in self.get_after_create_actions():
            result = action.fn(old_item)
            if isinstance(result, Awaitable):
                await result


def _get_triggered_actions(store_name: str, trigger_type) -> Iterator[Action]:
    for action, trigger in find_actions_with_trigger_type(trigger_type):
        if trigger.store_name == store_name:
            yield action



from dataclasses import dataclass, field
from typing import Optional

from marshy.types import ExternalItemType

from persisty.impl.mem.mem_store import MemStore
from persisty.store.restrict_access_store import restrict_access_store
from persisty.store.schema_validating_store import SchemaValidatingStore
from persisty.store.store_abc import StoreABC
from persisty.store_meta import StoreMeta
from persisty.trigger.wrapper import triggered_store


@dataclass
class MemStoreFactory:
    store_meta: StoreMeta
    items: ExternalItemType = field(default_factory=dict)

    def get_meta(self) -> StoreMeta:
        return self.store_meta

    def create(self) -> Optional[StoreABC]:
        store = MemStore(self.store_meta, self.items)
        store = SchemaValidatingStore(store)
        store = restrict_access_store(store, self.store_meta.store_access)
        store = triggered_store(store)
        return store



from __future__ import annotations
from typing import Tuple, Optional, Any, TYPE_CHECKING

from dataclasses import dataclass

from persisty.attr.attr_filter import AttrFilter, AttrFilterOp
from persisty.attr.attr_type import AttrType
from persisty.search_filter.or_filter import Or
from persisty.search_filter.search_filter_abc import SearchFilterABC, T

if TYPE_CHECKING:
    from persisty.attr.attr import Attr


@dataclass
class QueryFilter(SearchFilterABC[T]):
    query: str

    def __post_init__(self):
        object.__setattr__(self, "query", self.query.lower())

    def lock_attrs(self, attrs: Tuple[Attr, ...]) -> SearchFilterABC:
        filters = []
        for attr in attrs:
            if attr.readable and attr.type is AttrType.STR:
                filters.append(AttrFilter(attr.name, AttrFilterOp.contains, self.query))
        return Or(tuple(filters))

    def match(self, item: T, attrs: Tuple[Attr, ...]) -> bool:
        for attr in attrs:
            if not attr.readable or attr.type is not AttrType.STR:
                continue
            attr_value = item.get(attr.name)
            if isinstance(attr_value, str):
                if self.query in attr_value.lower():
                    return True
        return False

    def build_filter_expression(
        self, attrs: Tuple[Attr, ...]
    ) -> Tuple[Optional[Any], bool]:
        conditions = []
        for attr in attrs:
            if not attr.readable or attr.type is not AttrType.STR:
                continue
            from boto3.dynamodb.conditions import Attr

            conditions.append(Attr(attr.name).contains(self.query))
        if not conditions:
            return None, False
        if len(conditions) == 1:
            condition = conditions[0]
        else:
            from boto3.dynamodb.conditions import Or as DynOr

            condition = DynOr(*conditions)
        return condition, True



from typing import Any, Tuple, Iterator, Generic, TypeVar

from dataclasses import dataclass

from persisty.attr.attr import Attr
from persisty.util.undefined import UNDEFINED

T = TypeVar("T")


@dataclass(frozen=True)
class SearchOrderAttr(Generic[T]):
    attr: str
    desc: bool = False

    def validate_for_attrs(self, attrs: Tuple[Attr, ...]):
        for a in attrs:
            if a.name == self.attr and a.sortable:
                return
        raise ValueError(f"search_order_invalid:{self.attr}")

    def sort(self, items: Iterator[T]) -> Iterator[T]:
        items = sorted(items, key=self.key, reverse=self.desc)
        return items

    def key(self, item: T) -> Any:
        value = getattr(item, self.attr, UNDEFINED)
        if value in (None, UNDEFINED):
            return not self.desc, ""
        return self.desc, value

    def lt(self, a: T, b: T) -> bool:
        if self.desc:
            return self.key(a) > self.key(b)
        return self.key(a) < self.key(b)

    def eq(self, a: T, b: T) -> bool:
        return self.key(a) == self.key(b)



import os
from dataclasses import dataclass, field
from typing import Optional

import boto3
from opensearchpy import OpenSearch, RequestsHttpConnection, AWSV4SignerAuth

from persisty.search_filter.include_all import INCLUDE_ALL
from persisty.search_filter.search_filter_abc import SearchFilterABC
from persisty.store.store_abc import StoreABC
from persisty.store_meta import T, StoreMeta


@dataclass
class OpenSearchDocStore(StoreABC):
    meta: StoreMeta
    host: str
    port: str = 443
    region: str = field(default_factory=lambda: os.environ.get("AWS_REGION"))
    _open_search: Optional[OpenSearch] = None

    def get_meta(self) -> StoreMeta:
        return self.meta

    def create(self, item: T) -> Optional[T]:
        pass

    def read(self, key: str) -> Optional[T]:
        pass

    def _update(self, key: str, item: T, updates: T) -> Optional[T]:
        pass

    def _delete(self, key: str, item: T) -> bool:
        pass

    def count(self, search_filter: SearchFilterABC[T] = INCLUDE_ALL) -> int:
        pass

    @property
    def open_search(self) -> OpenSearch:
        if self._open_search:
            return self._open_search
        session = boto3.Session()
        credentials = session.get_credentials().get_frozen_credentials()
        auth = AWSV4SignerAuth(credentials, self.region)
        open_search = self._open_search = OpenSearch(
            hosts=[{"host": self.host, "port": self.port}],
            http_auth=auth,
            use_ssl=True,
            verify_certs=True,
            connection_class=RequestsHttpConnection,
        )
        return open_search



from dataclasses import dataclass
from datetime import datetime
from typing import Iterator

import marshy
from marshy.types import ExternalItemType, ExternalType

from persisty.attr.attr_type import AttrType
from persisty.errors import PersistyError
from persisty.key_config.key_config_abc import KeyConfigABC, T
from persisty.util import UNDEFINED


@dataclass(frozen=True)
class AttrKeyConfig(KeyConfigABC[T]):
    attr_name: str = "id"
    attr_type: AttrType = AttrType.UUID

    def to_key_str(self, item: T) -> str:
        value = getattr(item, self.attr_name)
        if not isinstance(value, str):
            if value in (None, UNDEFINED):
                raise PersistyError("invalid_key")
            value = str(value)
        return value

    def from_key_str(self, key: str, target: T):
        setattr(target, self.attr_name, self.to_key_attr(key))

    def to_key_dict(self, key: str) -> ExternalItemType:
        result = {self.attr_name: self.to_key_attr(key)}
        return result

    def to_key_attr(self, key: str) -> ExternalType:
        if key not in (None, UNDEFINED):
            if self.attr_type is AttrType.INT:
                return int(key)
            if self.attr_type is AttrType.FLOAT:
                return float(key)
            if self.attr_type in (AttrType.UUID, AttrType.STR):
                return str(key)
            if self.attr_name is AttrType.DATETIME:
                if isinstance(key, datetime):
                    return marshy.dump(key)
                if isinstance(key, str):
                    return key
            if self.attr_name is AttrType.BOOL:
                return bool(key)
        raise PersistyError("invalid_type")

    def get_key_attrs(self) -> Iterator[str]:
        key_attrs = getattr(self, "_key_attrs", None)
        if not key_attrs:
            key_attrs = frozenset((self.attr_name,))
            object.__setattr__(self, "_key_attrs", key_attrs)
        return key_attrs


ATTR_KEY_CONFIG = AttrKeyConfig()



import os
from dataclasses import Field, MISSING
from enum import Enum
from typing import Optional, Tuple, Type, Dict

from marshy.factory.optional_marshaller_factory import get_optional_type
from schemey import SchemaContext, get_default_schema_context

from servey.cache_control.cache_control_abc import CacheControlABC
from servey.cache_control.secure_hash_cache_control import SecureHashCacheControl

from persisty.attr.attr import Attr, DEFAULT_PERMITTED_FILTER_OPS
from persisty.attr.attr_filter_op import TYPE_FILTER_OPS, SORTABLE_TYPES
from persisty.attr.attr_type import attr_type, AttrType
from persisty.attr.generator.default_value_generator import DefaultValueGenerator
from persisty.attr.generator.defaults import (
    get_default_generator_for_create,
    get_default_generator_for_update,
)
from persisty.errors import PersistyError
from persisty.index.index_abc import IndexABC
from persisty.key_config.attr_key_config import AttrKeyConfig
from persisty.key_config.key_config_abc import KeyConfigABC
from persisty.link.link_abc import LinkABC
from persisty.store_meta import StoreMeta
from persisty.util import to_snake_case
from persisty.util.undefined import UNDEFINED


def stored(
    cls=None,
    *,
    key_config: Optional[KeyConfigABC] = None,
    cache_control: Optional[CacheControlABC] = None,
    batch_size: int = 100,
    schema_context: Optional[SchemaContext] = None,
    indexes: Tuple[IndexABC, ...] = tuple(),
):
    """Decorator inspired by dataclasses, containing stored meta."""
    if schema_context is None:
        schema_context = get_default_schema_context()

    # pylint: disable=R0912,R0914,R0915
    def wrapper(cls_):
        nonlocal key_config, cache_control, batch_size, indexes
        links_by_name = {}
        attrs_by_name = {}
        key_config, cache_control, batch_size, indexes = _derive_args(
            cls_,
            key_config,
            cache_control,
            batch_size,
            indexes,
            attrs_by_name,
            links_by_name,
        )
        cls_dict = cls_.__dict__
        annotations = cls_dict.get("__annotations__") or {}
        for name, type_ in annotations.items():
            if name.startswith("__"):
                continue
            value = cls_dict.get(name, UNDEFINED)
            if isinstance(value, LinkABC):
                continue
            if isinstance(value, Attr):
                attrs_by_name[value.name] = value
                continue

            creatable = True
            updatable = True
            update_generator = None
            if isinstance(value, Field):
                if value.metadata.get("persisty"):
                    attrs_by_name[name] = value.metadata.get("persisty")
                    continue
                if value.default is not MISSING:
                    create_generator = DefaultValueGenerator(value.default)
                else:
                    creatable, create_generator = get_default_generator_for_create(
                        name, type_
                    )
                updatable, update_generator = get_default_generator_for_update(
                    name, type_
                )
                schema = value.metadata.get(
                    "schemey"
                ) or schema_context.schema_from_type(type_)
            else:
                schema = schema_context.schema_from_type(type_)
                if value is UNDEFINED:
                    creatable, create_generator = get_default_generator_for_create(
                        name, type_
                    )
                    updatable, update_generator = get_default_generator_for_update(
                        name, type_
                    )
                else:
                    create_generator = DefaultValueGenerator(value)

            db_type = attr_type(type_)
            permitted_filter_ops, sortable = _derive_filter_and_sort(type, db_type)
            attr = Attr(
                name=name,
                attr_type=db_type,
                schema=schema,
                creatable=creatable,
                updatable=updatable,
                sortable=sortable,
                create_generator=create_generator,
                update_generator=update_generator,
                permitted_filter_ops=permitted_filter_ops,
            )
            attrs_by_name[name] = attr

        for name, value in cls_dict.items():
            if isinstance(value, LinkABC):
                links_by_name[name] = value
                value.update_attrs(attrs_by_name)

        store_meta = StoreMeta(
            name=to_snake_case(cls_.__name__),
            attrs=tuple(attrs_by_name.values()),
            key_config=_derive_key_config(key_config, cls_, attrs_by_name),
            cache_control=cache_control,
            batch_size=batch_size,
            description=cls_.__doc__,
            links=tuple(links_by_name.values()),
            indexes=indexes,
        )
        result = store_meta.get_stored_dataclass()
        return result

    return wrapper if cls is None else wrapper(cls)


def _is_enum(type_: Type):
    try:
        return issubclass(get_optional_type(type_), Enum)
    except TypeError:
        return False


def _derive_key_config(
    key_config: Optional[KeyConfigABC], cls: Type, attrs_by_name: Dict[str, Attr]
):
    if key_config:
        # Check that all attributes required by the key actually exist...
        for attr_name in key_config.get_key_attrs():
            if attr_name not in attrs_by_name:
                raise PersistyError(f"invalid_key_attr:{attr_name}")
        return key_config
    key_attr = attrs_by_name.get("id") or attrs_by_name.get("key")
    if not key_attr:
        raise PersistyError(f"could_not_derive_key:{cls}")
    return AttrKeyConfig(key_attr.name, key_attr.attr_type)


def _derive_filter_and_sort(type_: Type, db_type: AttrType):
    if _is_enum(type_):
        return DEFAULT_PERMITTED_FILTER_OPS, False
    return (TYPE_FILTER_OPS.get(db_type) or DEFAULT_PERMITTED_FILTER_OPS), (
        os.environ.get("PERSISTY_ATTRS_SORTABLE") != "0" and db_type in SORTABLE_TYPES
    )


# pylint: disable=R0913
def _derive_args(
    cls_: Type,
    key_config: Optional[KeyConfigABC],
    cache_control: Optional[CacheControlABC],
    batch_size: int,
    indexes: Tuple[IndexABC, ...],
    attrs_by_name: Dict[str, Attr],
    links_by_name: Dict[str, IndexABC],
):
    mro = list(cls_.__mro__)[1:-1]
    mro.reverse()
    for c in mro:
        store_meta = c.__dict__.get("__persisty_store_meta__")
        if store_meta:
            if key_config is None:
                key_config = store_meta.key_config
            if cache_control is None:
                cache_control = store_meta.cache_control
            if batch_size == 100:
                batch_size = store_meta.batch_size
            if indexes is None:
                indexes = store_meta.indexes
            links_by_name.update({link.get_name(): link for link in store_meta.links})
            attrs_by_name.update({attr.name: attr for attr in store_meta.attrs})
    if cache_control is None:
        cache_control = SecureHashCacheControl()
    return key_config, cache_control, batch_size, indexes



from dataclasses import dataclass
from typing import Tuple, Iterator

from marshy.types import ExternalItemType

from persisty.key_config.attr_key_config import AttrKeyConfig, T
from persisty.key_config.key_config_abc import KeyConfigABC
from persisty.util import to_base64, from_base64


@dataclass(frozen=True)
class CompositeKeyConfig(KeyConfigABC[T]):
    attrs: Tuple[AttrKeyConfig, ...]

    def to_key_str(self, item: T) -> str:
        keys = [a.to_key_str(item) for a in self.attrs]
        key = to_base64(keys)
        return key

    def from_key_str(self, key: str, target: T):
        result = from_base64(key)
        for attr, value in zip(self.attrs, result):
            setattr(target, attr.attr_name, value)

    def to_key_dict(self, key: str) -> ExternalItemType:
        key_parts = from_base64(key)
        result = {
            attr.attr_name: attr.to_key_attr(value)
            for attr, value in zip(self.attrs, key_parts)
        }
        return result

    def get_key_attrs(self) -> Iterator[str]:
        key_attrs = getattr(self, "_key_attrs", None)
        if not key_attrs:
            key_attrs = frozenset(a.attr_name for a in self.attrs)
            object.__setattr__(self, "_key_attrs", key_attrs)
        return key_attrs



from marshy.types import ExternalItemType
from servey.servey_aws.serverless.yml_config.yml_config_abc import (
    YmlConfigABC,
    ensure_ref_in_file,
    create_yml_file,
)

from persisty.finder.store_finder_abc import find_stores
from persisty.impl.default_store import DefaultStore
from persisty.impl.dynamodb.dynamodb_store_factory import DynamodbStoreFactory
from persisty.impl.dynamodb.dynamodb_table_store import DynamodbTableStore


class DynamodbYmlConfig(YmlConfigABC):
    """
    Set up some aspect of the serverless environment yml files. (For example, functions, resources, etc...)
    """

    dynamodb_resource_yml_file: str = "serverless_servey/dynamodb_resource.yml"
    dynamodb_role_statement_yml_file: str = (
        "serverless_servey/dynamodb_role_statement.yml"
    )

    def configure(self, main_serverless_yml_file: str):
        ensure_ref_in_file(
            main_serverless_yml_file,
            ["resources"],
            self.dynamodb_resource_yml_file,
        )
        ensure_ref_in_file(
            main_serverless_yml_file,
            ["provider", "iamRoleStatements"],
            self.dynamodb_role_statement_yml_file,
            "iamRoleStatements.0",
        )
        dynamodb_resource_yml = self.build_dynamodb_resource_yml()
        create_yml_file(self.dynamodb_resource_yml_file, dynamodb_resource_yml)
        dynamodb_role_statement_yml = self.build_dynamodb_role_statement_yml()
        create_yml_file(
            self.dynamodb_role_statement_yml_file, dynamodb_role_statement_yml
        )

    @staticmethod
    def get_dynamodb_store_factories():
        stores = find_stores()
        for store in stores:
            if isinstance(store, (DynamodbTableStore, DefaultStore)):
                factory = DynamodbStoreFactory(store.meta)
                factory.derive_from_meta()
                yield factory

    def build_dynamodb_resource_yml(self) -> ExternalItemType:
        resources = {}
        for factory in self.get_dynamodb_store_factories():
            resources[factory.table_name.title().replace("_", "")] = {
                "Type": "AWS::DynamoDB::Table",
                "Properties": {
                    "TableName": factory.table_name,
                    "AttributeDefinitions": factory.get_attribute_definitions(),
                    "KeySchema": factory.index.to_schema(),
                    "GlobalSecondaryIndexes": factory.get_global_secondary_indexes(),
                    "BillingMode": "PAY_PER_REQUEST",
                },
            }
        return {"Resources": resources}

    def build_dynamodb_role_statement_yml(self) -> ExternalItemType:
        resources = []
        for factory in self.get_dynamodb_store_factories():
            resource_name = factory.table_name.title().replace("_", "")
            resources.append({"Fn::GetAtt": [resource_name, "Arn"]})
            for index_name in factory.global_secondary_indexes.keys():
                resources.append(
                    {
                        "Fn::Join": [
                            "/",
                            [
                                {"Fn::GetAtt": [resource_name, "Arn"]},
                                "index",
                                index_name,
                            ],
                        ]
                    }
                )
        return {"iamRoleStatements": [self._iam_role_statement(resources)]}

    @staticmethod
    def _iam_role_statement(resource):
        result = {
            "Effect": "Allow",
            "Action": [
                "dynamodb:DescribeTable",
                "dynamodb:Query",
                "dynamodb:Scan",
                "dynamodb:BatchGetItem",
                "dynamodb:GetItem",
                "dynamodb:PutItem",
                "dynamodb:UpdateItem",
                "dynamodb:DeleteItem",
            ],
            "Resource": resource,
        }
        return result



from typing import Tuple, Iterator, Generic

from dataclasses import dataclass

from persisty.attr.attr import Attr
from persisty.search_order.search_order_attr import SearchOrderAttr, T


@dataclass(frozen=True)
class SearchOrder(Generic[T]):
    orders: Tuple[SearchOrderAttr[T], ...]

    def validate_for_attrs(self, attrs: Tuple[Attr, ...]):
        for order in self.orders:
            order.validate_for_attrs(attrs)

    def sort(self, items: Iterator[T]) -> Iterator[T]:
        for order in reversed(self.orders):
            items = order.sort(items)
        return items

    def lt(self, a: T, b: T) -> bool:
        for order in self.orders:
            if order.lt(a, b):
                return True
            if not order.eq(a, b):
                return False

    def eq(self, a: T, b: T) -> bool:
        for order in self.orders:
            if not order.eq(a, b):
                return False



import dataclasses
import inspect
from typing import Type, Optional, List

from servey.action.action import Action, action, get_action
from servey.action.batch_invoker import BatchInvoker
from servey.security.authorization import Authorization
from servey.trigger.web_trigger import WebTrigger, WebTriggerMethod

from persisty.batch_edit import batch_edit_dataclass_for
from persisty.batch_edit_result import batch_edit_result_dataclass_for
from persisty.factory.store_factory_abc import StoreFactoryABC
from persisty.link.link_abc import LinkABC
from persisty.result_set import result_set_dataclass_for
from persisty.search_filter.include_all import INCLUDE_ALL
from persisty.search_filter.search_filter_factory import SearchFilterFactoryABC
from persisty.search_order.search_order_factory import SearchOrderFactoryABC
from persisty.servey import generated
from persisty.store_meta import get_meta, StoreMeta


def action_for_create(
    store_factory: StoreFactoryABC,
    item_type: Type,
    create_input_type: Type,
) -> Action:
    store_meta = store_factory.get_meta()

    @action(
        name=f"{store_meta.name}_create",
        description=f"Create and return an item in {store_meta.name}",
        triggers=WebTrigger(
            WebTriggerMethod.POST, "/actions/" + store_meta.name.replace("_", "-")
        ),
    )
    def create(
        item: create_input_type, authorization: Optional[Authorization] = None
    ) -> Optional[item_type]:
        store = store_factory.create(authorization)
        created = store.create(item)
        return created

    return get_action(create)


def action_for_read(store_factory: StoreFactoryABC, item_type: Type) -> Action:
    store_meta = store_factory.get_meta()

    @action(
        name=f"{store_meta.name}_read",
        description=f"Read an item from {store_meta.name} given a key",
        triggers=(
            WebTrigger(
                WebTriggerMethod.GET,
                "/actions/" + store_meta.name.replace("_", "-") + "/{key}",
            ),
        ),
        cache_control=store_meta.cache_control,
    )
    def read(
        key: str, authorization: Optional[Authorization] = None
    ) -> Optional[item_type]:
        store = store_factory.create(authorization)
        result = store.read(key)
        return result

    return get_action(read)


def action_for_update(
    store_factory: StoreFactoryABC,
    item_type: Type,
    update_input_type: Type,
    search_filter_type: Type[SearchFilterFactoryABC],
) -> Action:
    store_meta = store_factory.get_meta()

    @action(
        name=f"{store_meta.name}_update",
        description=f"Update and return an item in {store_meta.name}",
        triggers=(
            WebTrigger(
                WebTriggerMethod.PATCH,
                "/actions/" + store_meta.name.replace("_", "-") + "/{key}",
            ),
        ),
    )
    def update(
        key: str,
        item: update_input_type,
        precondition: Optional[search_filter_type] = None,
        authorization: Optional[Authorization] = None,
    ) -> Optional[item_type]:
        store_meta.key_config.from_key_str(key, item)
        store = store_factory.create(authorization)
        search_filter = _create_search_filter(search_filter_type, precondition)
        updated = store.update(item, search_filter)
        return updated

    return get_action(update)


def action_for_delete(store_factory: StoreFactoryABC) -> Action:
    store_meta = store_factory.get_meta()

    @action(
        name=f"{store_meta.name}_delete",
        description=f"Delete an item in {store_meta.name}",
        triggers=(
            WebTrigger(
                WebTriggerMethod.DELETE,
                "/actions/" + store_meta.name.replace("_", "-") + "/{key}",
            ),
        ),
    )
    def delete(key: str, authorization: Optional[Authorization] = None) -> bool:
        store = store_factory.create(authorization)
        result = store.delete(key)
        return result

    return get_action(delete)


def action_for_search(
    store_factory: StoreFactoryABC,
    item_type: Type,
    search_filter_type: Optional[Type[SearchFilterFactoryABC]],
    search_order_type: Type[SearchOrderFactoryABC],
) -> Action:
    store_meta = store_factory.get_meta()
    # noinspection PyTypeChecker
    result_set_type = result_set_dataclass_for(item_type)
    setattr(generated, result_set_type.__name__, result_set_type)

    @action(
        name=f"{store_meta.name}_search",
        description=f"Run a search in {store_meta.name}",
        triggers=(
            WebTrigger(
                WebTriggerMethod.GET,
                f"/actions/{store_meta.name.replace('_', '-')}-search",
            ),
        ),
    )
    def search(
        search_filter: Optional[search_filter_type] = None,
        search_order: Optional[search_order_type] = None,
        page_key: Optional[str] = None,
        limit: Optional[int] = None,
        authorization: Optional[Authorization] = None,
    ) -> result_set_type:
        store = store_factory.create(authorization)
        search_filter = _create_search_filter(search_filter_type, search_filter)
        if search_order:
            # noinspection PyArgumentList,PyDataclass
            search_order = search_order_type(
                **{
                    f.name: getattr(search_order, f.name)
                    for f in dataclasses.fields(search_order_type)
                }
            )
            search_order = search_order.to_search_order()
        result_set = store.search(search_filter, search_order, page_key, limit)

        # noinspection PyArgumentList
        result_set = result_set_type(
            results=result_set.results,
            next_page_key=result_set.next_page_key,
        )
        return result_set

    if search_order_type is None:
        sig = inspect.signature(search)
        sig = sig.replace(
            parameters=[p for p in sig.parameters.values() if p.name != "search_order"]
        )
        search.__signature__ = sig

    search = action(
        fn=search,
        name=f"{store_meta.name}_search",
        description=f"Run a search in {store_meta.name}",
        triggers=(
            WebTrigger(
                WebTriggerMethod.GET,
                f"/actions/{store_meta.name.replace('_', '-')}-search",
            ),
        ),
    )

    return get_action(search)


def action_for_count(
    store_factory: StoreFactoryABC,
    search_filter_type: Type[SearchFilterFactoryABC],
) -> Action:
    store_meta = store_factory.get_meta()

    @action(
        name=f"{store_meta.name}_count",
        description=f"Get a count from {store_meta.name}",
        triggers=(
            WebTrigger(
                WebTriggerMethod.GET,
                f"/actions/{store_meta.name.replace('_', '-')}-count",
            ),
        ),
    )
    def count(
        search_filter: Optional[search_filter_type] = None,
        authorization: Optional[Authorization] = None,
    ) -> int:
        store = store_factory.create(authorization)
        search_filter = _create_search_filter(search_filter_type, search_filter)
        result = store.count(search_filter)
        return result

    return get_action(count)


def action_for_read_batch(store_factory: StoreFactoryABC, item_type: Type) -> Action:
    store_meta = store_factory.get_meta()

    @action(
        name=f"{store_meta.name}_read_batch",
        description=f"Read a batch of items from {store_meta.name} given keys",
        triggers=WebTrigger(
            WebTriggerMethod.GET,
            "/actions/" + store_meta.name.replace("_", "-") + "-batch",
        ),
        cache_control=store_meta.cache_control,
    )
    def read_batch(
        keys: List[str], authorization: Optional[Authorization] = None
    ) -> List[Optional[item_type]]:
        store = store_factory.create(authorization)
        results = store.read_batch(keys)
        return results

    return get_action(read_batch)


def action_for_edit_batch(
    store_factory: StoreFactoryABC,
    create_input_type: Type,
    update_input_type: Type,
) -> Action:
    store_meta = store_factory.get_meta()
    batch_edit_type = batch_edit_dataclass_for(
        store_meta.name.title() + "BatchEdit", create_input_type, update_input_type
    )
    batch_edit_result_type = batch_edit_result_dataclass_for(batch_edit_type)

    @action(
        name=f"{store_meta.name}_edit_batch",
        description=f"Perform a batch of edits against {store_meta.name}",
        triggers=WebTrigger(
            WebTriggerMethod.PATCH,
            "/actions/" + store_meta.name.replace("_", "-") + "-batch",
        ),
    )
    def edit_batch(
        edits: List[batch_edit_type], authorization: Optional[Authorization] = None
    ) -> List[batch_edit_result_type]:
        store = store_factory.create(authorization)
        results = store.edit_batch(edits)
        results = [
            batch_edit_result_type(
                edit=edit,
                success=result.success,
                code=result.code,
                details=result.details,
            )
            for edit, result in zip(edits, results)
        ]
        return results

    return get_action(edit_batch)


def wrap_links_in_actions(read_type: Type):
    meta = get_meta(read_type)
    overrides = {}
    for k, v in read_type.__dict__.items():
        if isinstance(v, LinkABC):
            overrides[k] = _to_action_fn(meta, v)
    if overrides:
        overrides["__doc__"] = read_type.__dict__.get("__doc__")
        read_type = dataclasses.dataclass(
            type(read_type.__name__, (read_type,), overrides)
        )
    setattr(generated, read_type.__name__, read_type)
    return read_type


def _to_action_fn(meta: StoreMeta, link: LinkABC):
    return_type = link.get_linked_type("persisty.servey.generated")

    # pylint: disable=C2801
    def wrapper(self, authorization: Optional[Authorization] = None) -> return_type:
        fn = link
        if hasattr(link, "__get__"):
            fn = link.__get__(self, self.__class__)
        # noinspection PyCallingNonCallable
        result = fn(authorization)
        return result

    batch_invoker = None
    if hasattr(link, "batch_call"):
        if hasattr(link, "arg_extractor"):
            arg_extractor = link.arg_extractor
        else:

            def arg_extractor(c):
                return [meta.key_config.to_key_str(c)]

        batch_invoker = BatchInvoker(
            fn=getattr(link, "batch_call"),
            arg_extractor=arg_extractor,
            max_batch_size=meta.batch_size,
        )

    wrapped = action(
        wrapper, name=meta.name + "_" + link.get_name(), batch_invoker=batch_invoker
    )

    return wrapped


def _create_search_filter(search_filter_factory_type, search_filter_input):
    # Since this may be from graphql, it may not actually be the class we expect, so we
    # replace the instance
    if not search_filter_input:
        return INCLUDE_ALL
    kwargs = {
        f.name: getattr(search_filter_input, f.name)
        for f in dataclasses.fields(search_filter_input)
    }
    search_filter_factory = search_filter_factory_type(**kwargs)
    search_filter = search_filter_factory.to_search_filter()
    return search_filter



from typing import Optional, Tuple, Any

from sqlalchemy import Table

from persisty.impl.sqlalchemy.search_filter.search_filter_converter_abc import (
    SearchFilterConverterABC,
)
from persisty.search_filter.search_filter_abc import SearchFilterABC
from persisty.store_meta import StoreMeta


class IncludeAllConverter(SearchFilterConverterABC):
    def convert(
        self,
        search_filter: SearchFilterABC,
        table: Table,
        store_meta: StoreMeta,
        context,
    ) -> Optional[Tuple[Any, bool]]:
        if search_filter is IncludeAllConverter:
            return None, True



from abc import ABC
from dataclasses import dataclass
from typing import Optional, ForwardRef, Union, Type

import typing_inspect
from servey.security.authorization import Authorization

from persisty.finder.store_finder_abc import find_store_factories
from persisty.link.link_abc import LinkABC
from persisty.util import to_snake_case


@dataclass
class LinkedStoreABC(LinkABC, ABC):
    linked_store_type: Union[Type, str, ForwardRef]
    name: Optional[str] = None  # Allows None so __set_name__ can exist
    linked_store_name: Optional[str] = None

    def get_linked_store_name(self):
        linked_store_name = self.linked_store_name
        if not linked_store_name:
            linked_type = self.linked_store_type
            if isinstance(linked_type, ForwardRef):
                linked_type = typing_inspect.get_forward_arg(linked_type).split(".")[-1]
            elif isinstance(linked_type, type):
                linked_type = linked_type.__name__
            linked_store_name = to_snake_case(linked_type)
        return linked_store_name

    def get_linked_store_factory(self):
        linked_store_factory = getattr(self, "_linked_store_factory", None)
        if not linked_store_factory:
            linked_store_factory = next(
                f
                for f in find_store_factories()
                if f.get_meta().name == self.get_linked_store_name()
            )
            setattr(self, "_linked_store_factory", linked_store_factory)
        return linked_store_factory

    def get_linked_store(self, authorization: Optional[Authorization] = None):
        store = self.get_linked_store_factory().create(authorization)
        return store



from dataclasses import dataclass
from typing import Type, Optional

from persisty.attr.attr_filter import AttrFilter
from persisty.attr.attr_type import AttrType
from persisty.search_filter.include_all import INCLUDE_ALL
from persisty.search_filter.query_filter import QueryFilter
from persisty.search_filter.search_filter_abc import SearchFilterABC
from persisty.store_meta import StoreMeta
from persisty.util import to_camel_case


class SearchFilterFactoryABC:
    """
    A lot of formats rely on search filters having a simple flat structure. This facilitates that pattern
    """

    __persisty_store_meta__: StoreMeta

    def to_search_filter(self) -> SearchFilterABC:
        search_filter = INCLUDE_ALL
        query = getattr(self, "query", None)
        if query:
            search_filter = QueryFilter(query)
        for attr in self.__persisty_store_meta__.attrs:
            for op in attr.permitted_filter_ops:
                filter_name = f"{attr.name}__{op.name}"
                value = getattr(self, filter_name, None)
                if value is not None:
                    search_filter &= AttrFilter(attr.name, op, value)
        return search_filter


def search_filter_dataclass_for(
    storage_meta: StoreMeta,
) -> Optional[Type[SearchFilterFactoryABC]]:
    annotations = {}
    has_str = next(
        (True for f in storage_meta.attrs if f.attr_type == AttrType.STR), False
    )
    if has_str:
        annotations["query"] = Optional[str]
    for attr in storage_meta.attrs:
        attr_type = Optional[attr.schema.python_type]
        for op in attr.permitted_filter_ops:
            annotations[f"{attr.name}__{op.name}"] = attr_type
    if not annotations:
        return
    params = {k: None for k in annotations}
    params["__persisty_store_meta__"] = storage_meta
    params["__annotations__"] = annotations
    params["__doc__"] = f"Search Filter for {storage_meta.name}"
    name = f"{to_camel_case(storage_meta.name)}SearchFilter"
    type_ = dataclass(type(name, (SearchFilterFactoryABC,), params))
    return type_



from dataclasses import dataclass
from typing import Optional

from servey.security.authorization import Authorization

from persisty.factory.store_factory_abc import StoreFactoryABC
from persisty.store.owned_store import OwnedStore, meta_with_non_editable_subject_id
from persisty.store.store_abc import StoreABC
from persisty.store_meta import T, StoreMeta


@dataclass
class OwnedStoreFactory(StoreFactoryABC[T]):
    """
    Factory for stores where ownership of items is asserted by means of an attribute name. The value for the
    attribute is set from the Authorization subject_id, and special permission may be required for update / delete
    operations
    """

    store_factory: StoreFactoryABC[T]
    subject_id_attr_name: str = "subject_id"
    require_ownership_for_read: bool = False
    require_ownership_for_update: bool = True
    require_ownership_for_delete: bool = True

    def get_meta(self) -> StoreMeta:
        meta = getattr(self, "_meta", None)
        if meta is None:
            meta = meta_with_non_editable_subject_id(
                self.store_factory.get_meta(), self.subject_id_attr_name
            )
            setattr(self, "_meta", meta)
        return meta

    def create(self, authorization: Optional[Authorization]) -> Optional[StoreABC[T]]:
        store = self.store_factory.create(authorization)
        store = OwnedStore(
            store=store,
            authorization=authorization,
            subject_id_attr_name=self.subject_id_attr_name,
            require_ownership_for_read=self.require_ownership_for_read,
            require_ownership_for_delete=self.require_ownership_for_delete,
            require_ownership_for_update=self.require_ownership_for_update,
        )
        return store



from typing import Optional, Tuple

from dataclasses import dataclass

from persisty.errors import PersistyError
from persisty.search_filter.search_filter_abc import SearchFilterABC
from persisty.store.filtered_store_abc import FilteredStoreABC, T
from persisty.store.store_abc import StoreABC


@dataclass(frozen=True)
class FilteredStore(FilteredStoreABC[T]):
    store: StoreABC[T]
    search_filter: SearchFilterABC[T]

    def get_store(self) -> StoreABC:
        return self.store

    def filter_create(self, item: T) -> Optional[T]:
        if not self.search_filter.match(item, self.get_meta().attrs):
            raise PersistyError("create_forbidden")
        return item

    # noinspection PyUnusedLocal
    def filter_update(self, item: T, updates: T) -> T:
        # old_item has already been checked in read operation
        item = {**item, **updates}
        if not self.search_filter.match(item, self.get_meta().attrs):
            raise PersistyError("update_forbidden")
        return updates

    def filter_read(self, item: T) -> Optional[T]:
        if self.search_filter.match(item, self.get_meta().attrs):
            return item

    # noinspection PyUnusedLocal
    def allow_delete(self, item: T) -> bool:
        return self.search_filter.match(item, self.get_meta().attrs)

    def filter_search_filter(
        self, search_filter: SearchFilterABC
    ) -> Tuple[SearchFilterABC, bool]:
        return search_filter & self.search_filter, True



from uuid import uuid4

from persisty.attr.generator.attr_value_generator_abc import AttrValueGeneratorABC
from persisty.util.undefined import UNDEFINED


class UuidGenerator(AttrValueGeneratorABC):
    def transform(self, value, item):
        if value is UNDEFINED:
            return uuid4()
        return value



import json
import os
from itertools import islice
from os.path import exists
from pathlib import Path
from typing import Type, List, Iterator

import marshy
from marshy.types import ExternalItemType

from persisty.factory.store_factory_abc import StoreFactoryABC
from persisty.finder.store_finder_abc import find_stores
from persisty.batch_edit import BatchEdit
from persisty.impl.default_store import DefaultStore
from persisty.store.store_abc import StoreABC
from persisty.store_meta import StoreMeta


def export_all(directory: str):
    """
    Export all store to yml files
    """
    for store in find_stores():
        export_meta(directory, store)
        export_content(directory, store)


def export_meta(directory: str, store: StoreABC):
    """
    Export store to json files
    """
    store_meta = store.get_meta()
    path = Path(directory, store_meta.name, "meta.json")
    path.parent.mkdir(exist_ok=True, parents=True)
    meta = marshy.dump(store_meta)
    # pylint: disable=W1514
    with open(path, "w") as f:
        json.dump(meta, f)


def export_content(directory: str, store: StoreABC, page_size: int = 500):
    """
    Export store content to json files
    """
    store_meta = store.get_meta()
    path = Path(directory, store_meta.name)
    path.mkdir(exist_ok=True, parents=True)
    results = store.search_all()
    index = 1
    while True:
        batch = list(islice(results, page_size))
        if not batch:
            return
        path = Path(directory, store_meta.name, str(index) + ".json")
        # pylint: disable=W1514
        with open(path, "w") as f:
            json.dump(batch, f)
        index += 1


def import_all(
    directory: str, store_type: Type = DefaultStore
) -> List[StoreFactoryABC]:
    results = []
    for store_name in os.listdir(directory):
        store_factory = import_store(directory, store_name, store_type)
        results.append(store_factory)
        import_content(directory, store_factory)
    return results


def import_store(
    directory: str,
    store_name: str,
    store_type: Type = DefaultStore,
) -> StoreFactoryABC:
    store_meta = import_meta(directory, store_name)
    factory = store_type(store_meta)
    return factory


def import_meta(directory: str, store_name: str) -> StoreMeta:
    path = Path(directory, store_name, "meta.yml")
    # pylint: disable=W1514
    with open(path, "r") as f:
        meta = json.load(f)
    store_meta = marshy.load(StoreMeta, meta)
    return store_meta


def import_content(directory: str, store: StoreABC):
    store_meta = store.get_meta()
    directory = Path(directory, store_meta.name)
    to_key_str = store_meta.key_config.to_key_str
    existing_item_edits = (
        BatchEdit(delete_key=to_key_str(i)) for i in store.search_all()
    )
    store.edit_all(existing_item_edits)
    items = read_all_items(directory)
    store.edit_all((BatchEdit(create_item=i) for i in items))


def read_all_items(directory: Path) -> Iterator[ExternalItemType]:
    index = 1
    while True:
        file = Path(directory, str(index) + ".yml")
        if not exists(file):
            return
        # pylint: disable=W1514
        with open(file, "r") as f:
            # noinspection PyTypeChecker
            items = json.loads(f)
            yield from items
            index += 1



from dataclasses import dataclass
from typing import Optional

from marshy import get_default_context
from marshy.marshaller.marshaller_abc import MarshallerABC
from schemey import Schema, get_default_schema_context

from persisty.errors import PersistyError
from persisty.store.filtered_store_abc import FilteredStoreABC, T
from persisty.store.store_abc import StoreABC
from persisty.store_meta import StoreMeta


@dataclass(frozen=True)
class SchemaValidatingStore(FilteredStoreABC[T]):
    store: StoreABC[T]
    schema_for_create: Schema = None
    schema_for_update: Schema = None
    marshaller_for_create: MarshallerABC[T] = None
    marshaller_for_update: MarshallerABC[T] = None

    def __post_init__(self):
        if not self.schema_for_create:
            object.__setattr__(
                self,
                "schema_for_create",
                get_default_schema_context().schema_from_type(
                    self.get_meta().get_create_dataclass()
                ),
            )
        if not self.schema_for_update:
            object.__setattr__(
                self,
                "schema_for_update",
                get_default_schema_context().schema_from_type(
                    self.get_meta().get_update_dataclass()
                ),
            )
        if not self.marshaller_for_create:
            object.__setattr__(
                self,
                "marshaller_for_create",
                get_default_context().get_marshaller(
                    self.get_meta().get_create_dataclass()
                ),
            )
        if not self.marshaller_for_update:
            object.__setattr__(
                self,
                "marshaller_for_update",
                get_default_context().get_marshaller(
                    self.get_meta().get_update_dataclass()
                ),
            )

    def get_store(self) -> StoreABC:
        return self.store

    def get_meta(self) -> StoreMeta:
        return self.store.get_meta()

    def filter_create(self, item: T) -> Optional[T]:
        dumped = self.marshaller_for_create.dump(item)
        error = next(self.schema_for_create.iter_errors(dumped), None)
        if error:
            raise PersistyError(error)
        return item

    def filter_update(self, item: T, updates: T) -> T:
        new_item = {
            **self.marshaller_for_update.dump(item),
            **self.marshaller_for_update.dump(updates),
        }
        error = next(self.schema_for_update.iter_errors(new_item), None)
        if error:
            raise PersistyError(error)
        return updates



from operator import or_
from typing import Optional, Tuple, Any

from sqlalchemy import Table

from persisty.attr.attr_type import AttrType
from persisty.impl.sqlalchemy.search_filter.and_filter_converter import (
    AndFilterConverter,
)
from persisty.search_filter.query_filter import QueryFilter
from persisty.search_filter.search_filter_abc import SearchFilterABC
from persisty.store_meta import StoreMeta

_NULL = None


class QueryFilterConverter(AndFilterConverter):
    # pylint: disable=E1120
    def convert(
        self,
        search_filter: SearchFilterABC,
        table: Table,
        store_meta: StoreMeta,
        context,
    ) -> Optional[Tuple[Any, bool]]:
        if isinstance(search_filter, QueryFilter):
            sub_clauses = []
            for attr in store_meta.attrs:
                if attr.type is AttrType.STR and attr.is_readable:
                    col = table.columns.get(attr.name)
                    sub_clauses.append(col.contains(search_filter.query))
            return or_(*sub_clauses), True



from typing import Optional, Tuple, Any
from uuid import UUID

from sqlalchemy import Table, Column

from persisty.attr.attr_filter import AttrFilter, AttrFilterOp
from persisty.impl.sqlalchemy.search_filter.and_filter_converter import (
    AndFilterConverter,
)
from persisty.search_filter.search_filter_abc import SearchFilterABC
from persisty.store_meta import StoreMeta

_NULL = None


class AttrFilterConverter(AndFilterConverter):
    def convert(
        self,
        search_filter: SearchFilterABC,
        table: Table,
        store_meta: StoreMeta,
        context,
    ) -> Optional[Tuple[Any, bool]]:
        if isinstance(search_filter, AttrFilter):
            col = table.columns.get(search_filter.name)
            clause = self.get_clause(col, search_filter.op, search_filter.value)
            return clause, True

    # pylint: disable=R0911,R0912
    @staticmethod
    def get_clause(col: Column, op: AttrFilterOp, value: Any):
        """
        Since we are building sqlalchemy where clauses here - the syntax is a little misleading. col == value
        does not do a direct comparison, but rather produces a where clause for comparing the contents of a column
        to the value given.
        """
        if isinstance(value, UUID):
            value = str(value)
        if op == AttrFilterOp.contains:
            return col.contains(value)
        if op == AttrFilterOp.endswith:
            return col.endswith(value)
        if op == AttrFilterOp.eq:
            return col == value
        if op == AttrFilterOp.exists:
            return col != _NULL  # yields col IS NOT NULL
        if op == AttrFilterOp.gt:
            return col > value
        if op == AttrFilterOp.gte:
            return col >= value
        if op == AttrFilterOp.lt:
            return col < value
        if op == AttrFilterOp.lte:
            return col <= value
        if op == AttrFilterOp.ne:
            return col != value
        if op == AttrFilterOp.not_exists:
            return col == _NULL  # yields col IS NULL
        if op == AttrFilterOp.oneof:
            return col.in_(*value)
        if op == AttrFilterOp.startswith:
            return col.startswith(value)



import json
from datetime import timezone
from typing import Optional, List, Iterator, Tuple, Any, Dict

from dataclasses import dataclass
from uuid import UUID

import marshy
from marshy.types import ExternalItemType
from sqlalchemy import Table, and_, select, func, Column
from sqlalchemy.engine import Engine
from sqlalchemy.exc import DatabaseError
from sqlalchemy.sql.elements import BindParameter, or_

from persisty.errors import PersistyError
from persisty.attr.attr_type import AttrType
from persisty.impl.sqlalchemy.search_filter.search_filter_converter_context import (
    SearchFilterConverterContext,
)
from persisty.impl.sqlalchemy.sqlalchemy_column_converter import POSTGRES
from persisty.search_filter.exclude_all import EXCLUDE_ALL
from persisty.batch_edit import BatchEdit
from persisty.batch_edit_result import BatchEditResult
from persisty.result_set import ResultSet
from persisty.search_filter.include_all import INCLUDE_ALL
from persisty.search_filter.search_filter_abc import SearchFilterABC
from persisty.search_order.search_order import SearchOrder
from persisty.store.store_abc import StoreABC, T
from persisty.store_meta import StoreMeta
from persisty.util import UNDEFINED, from_base64, to_base64


def catch_db_error(fn):
    def wrapper(*args, **kwargs):
        try:
            return fn(*args, **kwargs)
        except DatabaseError as e:
            raise PersistyError(e) from e

    return wrapper


@dataclass(frozen=True)
class SqlalchemyTableStore(StoreABC):
    """
    This class uses sql alchemy at a lower level than the standard orm usage
    """

    meta: StoreMeta
    table: Table
    engine: Engine

    def get_meta(self) -> StoreMeta:
        return self.meta

    @catch_db_error
    def create(self, item: T) -> Optional[T]:
        dumped = self._dump(item, False)
        with self.engine.begin() as connection:
            result = connection.execute(self.table.insert(), parameters=dumped)
            if result.inserted_primary_key:
                row = self._load_row(result.inserted_primary_key)
                for attr in self.meta.attrs:
                    value = getattr(row, attr.name)
                    if value is UNDEFINED:
                        setattr(row, attr.name, value)
            connection.commit()
            return item

    @catch_db_error
    def read(self, key: str) -> Optional[T]:
        with self.engine.begin() as connection:
            key_dict = self.meta.key_config.to_key_dict(key)
            return self._read(connection, key_dict)

    def _read(self, connection, key_dict: ExternalItemType) -> Optional[Dict]:
        stmt = self.table.select().where(self._key_where_clause())
        row = connection.execute(stmt, key_dict).first()
        if row:
            item = self._load_row(row)
            return item

    def read_batch(self, keys: List[str]) -> List[Optional[T]]:
        with self.engine.begin() as connection:
            key_config = self.meta.key_config
            key_objs = [key_config.to_key_dict(key) for key in keys]
            items = self._read_batch(connection, key_objs)
            items_by_key = {key_config.to_key_str(item): item for item in items}
            items = [items_by_key.get(k) for k in keys]
            return items

    def _read_batch(
        self, connection, keys: List[T], cols: Optional[List[Column]] = None
    ) -> List[Dict]:
        # NB: Does not enforce ordering
        assert len(keys) <= self.meta.batch_size
        where_clause = self._key_where_clause_from_dicts(keys)
        if cols:
            stmt = select(*cols)
        else:
            stmt = self.table.select()
        stmt = stmt.where(where_clause)
        results = connection.execute(stmt)
        items = [self._load_row(r) for r in results]
        return items

    # pylint: disable=W0613
    @catch_db_error
    def _update(
        self,
        key: str,
        item: T,
        updates: T,
        search_filter: SearchFilterABC = INCLUDE_ALL,
    ) -> Optional[T]:
        with self.engine.begin() as connection:
            stmt = self.table.update(self._key_where_clause_from_item(updates))
            updates = self._dump(updates, True)
            connection.execute(stmt, updates)
            key_dict = self.meta.key_config.to_key_dict(key)
            loaded = self._read(connection, key_dict)
            connection.commit()
            return loaded

    @catch_db_error
    def delete(self, key: str) -> bool:
        with self.engine.begin() as connection:
            key = self.meta.key_config.to_key_dict(key)
            stmt = self.table.delete(whereclause=self._key_where_clause())
            result = connection.execute(stmt, key)
            connection.commit()
            return bool(result.rowcount)

    @catch_db_error
    def _delete(self, key: str, item: T) -> bool:
        return self.delete(key)

    @catch_db_error
    def count(self, search_filter: SearchFilterABC = INCLUDE_ALL) -> int:
        if search_filter is EXCLUDE_ALL:
            return 0
        where_clause, handled = self._search_filter_to_where_clause(search_filter)
        if not handled:
            count = sum(1 for _ in self.search_all(search_filter))
            return count
        stmt = select([func.count()]).select_from(self.table)
        if where_clause is not None:
            stmt = stmt.where(where_clause)
        with self.engine.begin() as connection:
            row = connection.execute(stmt).first()
            return row[0]

    # pylint: disable=R0914,E1101
    @catch_db_error
    def search(
        self,
        search_filter: SearchFilterABC = INCLUDE_ALL,
        search_order: Optional[SearchOrder] = None,
        page_key: Optional[str] = None,
        limit: Optional[int] = None,
    ) -> ResultSet[T]:
        assert limit <= self.meta.batch_size
        where_clause, handled = self._search_filter_to_where_clause(search_filter)
        order_by = self._search_order_to_order_by(search_order)
        if order_by is None:
            order_by = self._default_order_by()
        offset = 0
        if page_key:
            page_key = from_base64(page_key)
            if page_key[0] == "offset":
                offset = page_key[1]
            else:
                # noinspection PyTypeChecker
                key_where_clause = and_(
                    self.table.columns.get(a) > page_key[1][a]
                    for a in self.meta.key_config.get_key_attrs()
                )
                where_clause = (
                    and_(where_clause, key_where_clause)
                    if where_clause is not None
                    else key_where_clause
                )

        stmt = self.table.select()
        if where_clause is not None:
            stmt = stmt.where(where_clause)
        stmt = stmt.order_by(*order_by)
        if offset:
            stmt = stmt.offset(offset)
        if handled:
            stmt = stmt.limit(limit)

        with self.engine.begin() as connection:
            rows = connection.execute(stmt)
            results = []
            next_page_key = None
            for row in rows:
                result = self._load_row(row)
                if handled or search_filter.match(result, self.meta.attrs):
                    results.append(result)
                    if len(results) == limit:
                        if search_order and search_order.orders:
                            next_page_key = ["offset", offset + len(results)]
                        if not search_order or not search_order.orders:
                            next_page_key = [
                                "id",
                                {
                                    a: _transform_type(getattr(result, a))
                                    for a in self.meta.key_config.get_key_attrs()
                                },
                            ]
                        next_page_key = to_base64(next_page_key)
                        break
            return ResultSet(results, next_page_key)

    @catch_db_error
    def search_all(
        self,
        search_filter: SearchFilterABC = INCLUDE_ALL,
        search_order: Optional[SearchOrder] = None,
    ) -> Iterator[T]:
        if search_filter is EXCLUDE_ALL:
            return ResultSet([])
        where_clause, handled = self._search_filter_to_where_clause(search_filter)
        order_by = self._search_order_to_order_by(search_order)
        stmt = self.table.select()
        if where_clause is not None:
            stmt = stmt.where(where_clause)
        if order_by is not None:
            stmt = stmt.order_by(*order_by)
        with self.engine.begin() as connection:
            rows = connection.execute(stmt)
            for row in rows:
                item = self._load_row(row)
                if not handled and not search_filter.match(item, self.meta.attrs):
                    continue
                yield item

    @catch_db_error
    def edit_batch(self, edits: List[BatchEdit]) -> List[BatchEditResult]:
        assert len(edits) <= self.meta.batch_size
        results_by_id = {}
        inserts = [e for e in edits if e.create_item]
        updates = [e for e in edits if e.update_item]
        deletes = [e for e in edits if e.delete_key]
        with self.engine.begin() as connection:
            if inserts:
                self._batch_insert(connection, inserts, results_by_id)
            if updates:
                self._batch_update(connection, updates, results_by_id)
            if deletes:
                self._batch_delete(connection, deletes, results_by_id)
            results = [results_by_id[e.id] for e in edits]
            connection.commit()
            return results

    def _batch_insert(
        self,
        connection,
        edits: List[BatchEdit],
        results_by_id: Dict[UUID, BatchEditResult],
    ):
        stmt = self.table.insert(bind=connection)
        items_to_create = [self._dump(e.create_item, False) for e in edits]
        connection.execute(stmt, items_to_create)
        for insert in edits:
            results_by_id[insert.id] = BatchEditResult(insert, True)

    # pylint: disable=E1101
    def _batch_update(
        self,
        connection,
        edits: List[BatchEdit],
        results_by_id: Dict[UUID, BatchEditResult],
    ):
        edits_by_key = {}
        key_dicts_to_load_row = []
        key_attrs = self.meta.key_config.get_key_attrs()
        to_key_str = self.meta.key_config.to_key_str
        for edit in edits:
            edit_key = to_key_str(edit.update_item)
            edits_by_key[edit_key] = edit
            key_dicts_to_load_row.append(
                {k: _transform_type(getattr(edit.update_item, k)) for k in key_attrs}
            )
            results_by_id[edit.id] = BatchEditResult(edit)
        existing_items = self._read_batch(connection, key_dicts_to_load_row)
        for item in existing_items:
            key = to_key_str(item)
            edit = edits_by_key[key]
            for attr in self.meta.attrs:
                value = UNDEFINED
                if attr.update_generator:
                    if attr.updatable:
                        value = attr.update_generator.transform(
                            getattr(edit.update_item, attr.name), edit.update_item
                        )
                    else:
                        value = attr.update_generator.transform(
                            UNDEFINED, edit.update_item
                        )
                elif attr.updatable:
                    value = getattr(edit.update_item, attr.name)
                    if value is UNDEFINED:
                        value = getattr(item, attr.name)
                if value is not UNDEFINED:
                    setattr(item, attr.name, _transform_type(value))
            item_updates = self._dump(item, True)
            stmt = self.table.update(self._key_where_clause_from_item(item))
            connection.execute(stmt, item_updates)
            results_by_id[edit.id].success = True

    def _batch_delete(
        self,
        connection,
        edits: List[BatchEdit],
        results_by_id: Dict[UUID, BatchEditResult],
    ):
        key_config = self.meta.key_config
        delete_keys = [key_config.to_key_dict(d.delete_key) for d in edits]
        existing_keys = self._get_existing_keys(connection, delete_keys)
        where_clause = self._key_where_clause_from_dicts(delete_keys)
        stmt = self.table.delete(whereclause=where_clause)
        connection.execute(stmt)
        deleted_keys = {key_config.to_key_str(k) for k in existing_keys}
        for delete in edits:
            deleted = delete.delete_key in deleted_keys
            results_by_id[delete.id] = BatchEditResult(delete, deleted)

    def _get_existing_keys(self, connection, keys: List[Dict]) -> List[Dict]:
        key_cols = [self.table.columns[a] for a in self.meta.key_config.get_key_attrs()]
        existing_keys = self._read_batch(connection, keys, key_cols)
        return existing_keys

    def _load_row(self, row):
        # Row is a KeyedTuple - _asdict is to match the namedtuple API (it's not private!)
        # noinspection PyProtectedMember
        item = row._asdict()
        return self._load(item)

    def _load(self, item: Dict):
        loaded = {}
        for attr_ in self.meta.attrs:
            if attr_.name not in item:
                continue
            value = item.get(attr_.name)
            if (
                attr_.attr_type == AttrType.JSON
                and not self.engine.dialect.name == POSTGRES
            ):
                value = json.loads(value)
            elif attr_.attr_type == AttrType.DATETIME and value:
                if not value.tzinfo:
                    value.replace(tzinfo=timezone.utc)
                value = value.strftime("%Y-%m-%dT%H:%M:%S+00:00")
            loaded[attr_.name] = value
        # noinspection PyTypeChecker
        result = marshy.load(self.meta.get_read_dataclass(), loaded)
        return result

    def _dump(self, item: T, is_update: bool):
        dumped = {}
        for attr_ in self.meta.attrs:
            value = getattr(item, attr_.name, UNDEFINED)
            if is_update:
                generator = attr_.update_generator
            else:
                generator = attr_.create_generator
            if generator:
                value = generator.transform(value, item)
            setattr(item, attr_.name, value)
            if value is UNDEFINED:
                continue
            if (
                attr_.attr_type == AttrType.JSON
                and not self.engine.dialect.name == POSTGRES
            ):
                value = json.dumps(value)
            dumped[attr_.name] = _transform_type(value)

        return dumped

    def _key_where_clause(self):
        key_where_clause = None
        for attr_name in self.meta.key_config.get_key_attrs():
            # pylint: disable=E1101
            exp = self.table.columns.get(attr_name) == BindParameter(attr_name)
            if key_where_clause:
                key_where_clause &= exp
            else:
                key_where_clause = exp
        return key_where_clause

    def _key_where_clause_from_dicts(self, dicts: List[ExternalItemType]):
        key_attrs = list(self.meta.key_config.get_key_attrs())
        if len(key_attrs) == 1:
            keys_for_where = [d.get(key_attrs[0]) for d in dicts]
            where_clause = self.table.columns[key_attrs[0]].in_(keys_for_where)
        else:
            where_clause = []
            for d in dicts:
                where_clause.append(self._key_where_clause_from_dict(d))
            where_clause = or_(*where_clause)
        return where_clause

    def _key_where_clause_from_dict(self, item: ExternalItemType):
        # pylint: disable=E1101
        where_clause = and_(
            self.table.columns.get(attr_name) == item.get(attr_name)
            for attr_name in self.meta.key_config.get_key_attrs()
        )
        return where_clause

    def _key_where_clause_from_item(self, item: T):
        where_clause = and_(
            self.table.columns.get(attr_name) == getattr(item, attr_name)
            for attr_name in self.meta.key_config.get_key_attrs()
        )
        return where_clause

    def _key_where_clause_from_items(self, items: List[T]):
        key_attrs = list(self.meta.key_config.get_key_attrs())
        if len(key_attrs) == 1:
            keys_for_where = [getattr(item, key_attrs[0]) for item in items]
            where_clause = self.table.columns[key_attrs[0]].in_(keys_for_where)
        else:
            where_clause = []
            for item in items:
                where_clause.append(self._key_where_clause_from_item(item))
            where_clause = or_(*where_clause)
        return where_clause

    def _search_filter_to_where_clause(
        self, search_filter: SearchFilterABC
    ) -> Tuple[Any, bool]:
        search_filter = search_filter.lock_attrs(self.meta.attrs)
        context = SearchFilterConverterContext()
        return context.convert(search_filter, self.table, self.meta)

    def _search_order_to_order_by(self, search_order: SearchOrder):
        if not search_order:
            return
        search_order.validate_for_attrs(self.meta.attrs)
        orders = []
        for order_attr in search_order.orders:
            # pylint: disable=E1101
            order = self.table.columns.get(order_attr.attr)
            if order_attr.desc:
                order = order.desc()
            orders.append(order)
        return orders

    def _default_order_by(self) -> List[Column]:
        # pylint: disable=E1101
        orders = [
            self.table.columns.get(attr_name)
            for attr_name in self.meta.key_config.get_key_attrs()
        ]
        return orders

    def _key_as_dict(self, item: T):
        result = self.meta.get_stored_dataclass()()
        required = self.meta.key_config.get_key_attrs()
        for attr in self.meta.attrs:
            if attr.name in required:
                setattr(result, attr.name, getattr(item, attr.name))
        return marshy.dump(result)


def _transform_type(value):
    if isinstance(value, UUID):
        return str(value)
    return value



from abc import ABC, abstractmethod
from typing import Dict, Type, Union, ForwardRef

from marshy.types import ExternalItemType

from persisty.attr.attr import Attr


class LinkABC(ABC):
    """
    A link represents something external to a stored item but related to it.
    """

    @abstractmethod
    def __set_name__(self, owner, name):
        """Called when a link is set as a value for a @stored entity."""

    @abstractmethod
    def get_name(self) -> str:
        """Get the name of this link"""

    def get_linked_type(self, forward_ref_ns: str) -> Union[ForwardRef, Type]:
        """Get the type for the linked object (May be an int, Optional[T], ResultSet[T], or something else"""

    def update_attrs(self, attrs_by_name: Dict[str, Attr]):
        """
        Update parameters to include anything required by this link that may be missing. (e.g. linked_id)
        """

    def update_json_schema(self, json_schema: ExternalItemType):
        """
        Update a schema with data from this link (Typically this is used as a custom extension to jsonschema to
        represent relational data.
        """



from dataclasses import dataclass
from typing import Optional, List, Dict, Tuple

from persisty.attr.attr_filter import AttrFilter
from persisty.attr.attr_filter_op import AttrFilterOp
from persisty.batch_edit import BatchEdit
from persisty.batch_edit_result import BatchEditResult
from persisty.errors import PersistyError
from persisty.index.unique_index import UniqueIndex
from persisty.search_filter.and_filter import And
from persisty.store.store_abc import StoreABC, T
from persisty.store.wrapper_store_abc import WrapperStoreABC
from persisty.util import UNDEFINED


@dataclass
class UniqueIndexStore(WrapperStoreABC[T]):
    store: StoreABC[T]
    unique_indexes: Tuple[UniqueIndex, ...]

    def get_store(self) -> StoreABC[T]:
        return self.store

    def create(self, item: T) -> T:
        self._check_create(item)
        return self.get_store().create(item)

    # pylint: disable=W0212
    def _update(self, key: str, item: T, updates: T) -> Optional[T]:
        self._check_update(key, item, updates)
        return self.get_store()._update(key, item, updates)

    # pylint: disable=W0212
    def _edit_batch(
        self, edits: List[BatchEdit[T, T]], items_by_key: Dict[str, T]
    ) -> List[BatchEditResult[T, T]]:
        filtered_edits = []
        key_config = self.store.get_meta().key_config
        for edit in edits:
            if edit.create_item:
                self._check_create(edit.create_item)
                filtered_edits.append(edit)
            elif edit.update_item:
                key = key_config.to_key_str(edit.update_item)
                item = items_by_key[key]
                updates = edit.update_item
                self._check_update(key, item, updates)
                filtered_edits.append(edit)
            else:
                filtered_edits.append(edit)
        return self.get_store()._edit_batch(edits, items_by_key)

    def _check_create(self, item: T):
        store = self.get_store()
        for index in self.unique_indexes:
            search_filter = And(
                tuple(
                    AttrFilter(a, AttrFilterOp.eq, getattr(item, a))
                    for a in index.attr_names
                )
            )
            item = next(store.search_all(search_filter), None)
            if item:
                raise PersistyError("non_unique_item")

    def _check_update(self, key: str, item: T, updates: T):
        store = self.get_store()
        key_config = store.get_meta().key_config
        for index in self.unique_indexes:
            search_filters = []
            for attr_name in index.attr_names:
                value = getattr(updates, attr_name)
                if value is UNDEFINED:
                    value = getattr(item, attr_name)
                search_filters.append(AttrFilter(attr_name, AttrFilterOp.eq, value))
            search_filter = And(search_filters)
            results = (
                item
                for item in store.search_all(search_filter)
                if key_config.to_key_str(item) != key
            )
            item = next(results, None)
            if item:
                raise PersistyError("non_unique_item")


def unique_index_store(store: StoreABC):
    meta = store.get_meta()
    unique_indexes = tuple(i for i in meta.indexes if isinstance(i, UniqueIndex))
    if unique_indexes:
        return UniqueIndexStore(store, unique_indexes)
    return store



import base64
import hashlib
from dataclasses import dataclass, field
from typing import Dict, Iterator, List, Tuple

from sqlalchemy import Table, MetaData, Index
from sqlalchemy.engine import Engine

from persisty.impl.sqlalchemy.sqlalchemy_column_converter import (
    SqlalchemyColumnConverter,
)
from persisty.impl.sqlalchemy.sqlalchemy_constraint_converter import (
    SqlalchemyConstraintConverter,
)
from persisty.index.attr_index import AttrIndex
from persisty.index.unique_index import UniqueIndex
from persisty.store_meta import StoreMeta


@dataclass
class SqlalchemyTableConverter:
    """Converter for storemeta to / from sqlalchemy tables"""

    engine: Engine
    metadata: MetaData
    schema: Dict[str, StoreMeta] = field(default_factory=dict)

    # pylint: disable=R0914
    def to_sqlalchemy_table_and_indexes(
        self, store_meta: StoreMeta
    ) -> Tuple[Table, List[Index]]:
        args = [
            store_meta.name,
            self.metadata,
        ]
        dialect = self.engine.dialect.name
        column_factory = SqlalchemyColumnConverter(store_meta.key_config, dialect)
        columns_by_name = {}
        indexes = []
        for attr_ in store_meta.attrs:
            column = column_factory.create_column(attr_)
            columns_by_name[attr_.name] = column
            args.append(column)
        for index in store_meta.indexes:
            if isinstance(index, AttrIndex):
                name = f"idx_{index.attr_name}"
                index_cols = [columns_by_name[index.attr_name]]
                unique = False
            elif isinstance(index, UniqueIndex):
                name = f"idx_{'__'.join(index.attr_names)}"
                index_cols = [columns_by_name[a] for a in index.attr_names]
                unique = True
            else:
                continue
            if len(name) > 62:
                name = base64.b64decode(hashlib.md5(name.encode("UTF-8"))).decode(
                    "UTF-8"
                )
            index_obj = Index(name, *index_cols, unique=unique)
            indexes.append(index_obj)

        constraint_factory = SqlalchemyConstraintConverter(self.schema)
        args.extend(constraint_factory.get_foreign_key_constraints(store_meta))
        table = Table(*args)
        return table, indexes

    def to_sqlalchemy_tables_and_indexes(self) -> Iterator[Tuple[Table, List[Index]]]:
        for store_meta in self.schema.values():
            table_and_indexes = self.to_sqlalchemy_table_and_indexes(store_meta)
            yield table_and_indexes



import asyncio
from dataclasses import dataclass
from typing import Optional, Iterator

from marshy.types import ExternalItemType
from servey.action.action import Action
from servey.finder.action_finder_abc import find_actions_with_trigger_type

from persisty.search_filter.include_all import INCLUDE_ALL
from persisty.search_filter.search_filter_abc import SearchFilterABC
from persisty.store.store_abc import StoreABC
from persisty.store.wrapper_store_abc import WrapperStoreABC
from persisty.trigger.store_triggers import StoreTriggers


@dataclass
class AsyncioTriggerStore(WrapperStoreABC):
    """
    Store which runs triggers after edits using asyncio
    """

    store: StoreABC
    store_triggers: StoreTriggers

    def get_store(self) -> StoreABC:
        return self.store

    def create(self, item: ExternalItemType) -> Optional[ExternalItemType]:
        result = self.store.create(item)
        if result:
            coro = self.store_triggers.async_after_create(result)
            asyncio.ensure_future(coro)
            return result

    def update(
        self, updates: ExternalItemType, precondition: SearchFilterABC = INCLUDE_ALL
    ) -> Optional[ExternalItemType]:
        if not self.store_triggers.has_after_update_actions():
            return self.store.update(updates, precondition)
        key = self.get_meta().key_config.to_key_str(updates)
        old_item = self.store.read(key)
        if old_item:
            new_item = self.store.update(updates, precondition)
            if new_item:
                coro = self.store_triggers.async_after_update(old_item, new_item)
                asyncio.ensure_future(coro)
                return new_item

    def delete(self, key: str) -> bool:
        if not self.store_triggers.has_after_delete_actions():
            return self.store.delete(key)
        result = False
        item = self.store.read(key)
        if item:
            result = self.store.delete(key)
            if result:
                coro = self.store_triggers.async_after_delete(item)
                asyncio.ensure_future(coro)
        return result


def _get_triggered_actions(store_name: str, trigger_type) -> Iterator[Action]:
    for action, trigger in find_actions_with_trigger_type(trigger_type):
        if trigger.store_name == store_name:
            yield action



import importlib
import logging
import os
import pkgutil
from dataclasses import dataclass, field
from typing import Iterator

from servey.util import get_servey_main

from persisty.factory.store_factory_abc import StoreFactoryABC
from persisty.finder.store_finder_abc import StoreFactoryFinderABC
from persisty.store.store_abc import StoreABC

LOGGER = logging.getLogger(__name__)


@dataclass
class ModuleStoreFinder(StoreFactoryFinderABC):
    """
    Default implementation of store factory finder which searches for actions in a particular module
    """

    root_module_name: str = field(
        default_factory=lambda: f"{os.environ.get('PERSISTY_MAIN') or get_servey_main()}.store"
    )

    def find_stores(self) -> Iterator[StoreABC]:
        module = importlib.import_module(self.root_module_name)
        # noinspection PyTypeChecker
        yield from find_instances_in_module(module, StoreABC)

    def find_store_factories(self) -> Iterator[StoreFactoryABC]:
        module = importlib.import_module(self.root_module_name)
        # noinspection PyTypeChecker
        yield from find_instances_in_module(module, StoreFactoryABC)


def find_instances_in_module(module, type_) -> Iterator:
    for value in list(module.__dict__.values()):
        if isinstance(value, type_):
            yield value
    if not hasattr(module, "__path__"):
        return  # Module was not a package...
    paths = []
    paths.extend(module.__path__)
    module_infos = list(pkgutil.walk_packages(paths))
    for module_info in module_infos:
        sub_module_name = module.__name__ + "." + module_info.name
        sub_module = importlib.import_module(sub_module_name)
        # noinspection PyTypeChecker
        yield from find_instances_in_module(sub_module, type_)



import base64
import hashlib
import json
import logging
import os
import re
import sys

from dataclasses import fields
from marshy.types import ExternalType, ExternalItemType

from persisty.util.undefined import UNDEFINED

_PATTERN = re.compile(r"(?<!^)(?=[A-Z])")


def to_base64(item: ExternalType) -> str:
    json_str = json.dumps(item)
    json_bytes = json_str.encode("utf-8")
    base64_bytes = base64.b64encode(json_bytes)
    base64_str = base64_bytes.decode("utf-8")
    return base64_str


def from_base64(base64_str: str) -> ExternalType:
    base64_bytes = base64_str.encode("utf-8")
    json_bytes = base64.b64decode(base64_bytes)
    json_str = json_bytes.decode("utf-8")
    item = json.loads(json_str)
    return item


def secure_hash(item) -> str:
    item_json = json.dumps(item)
    item_bytes = item_json.encode("utf-8")
    sha = hashlib.sha256()
    sha.update(item_bytes)
    hash_bytes = sha.digest()
    b64_bytes = base64.b64encode(hash_bytes)
    b64_str = b64_bytes.decode("utf-8")
    return b64_str


def filter_none(item: ExternalItemType) -> ExternalItemType:
    return {k: v for k, v in item.items() if v is not None}


LEVEL = os.environ.get("LOG_LEVEL", "INFO").upper()


def get_logger(name: str):
    """
    Get a logger with it's level set from the default environment variable for logging
    :param name: the name for the logger
    :return: a logger
    """
    logger = logging.getLogger(name)
    logger.addHandler(logging.StreamHandler(sys.stdout))
    logger.setLevel(LEVEL)
    return logger


def dataclass_to_params(dataclass):
    """Convert a dataclass to a dict, skipping any values which are UNDEFINED"""
    items = {
        field.name: getattr(dataclass, field.name)
        for field in fields(dataclass)
        if getattr(dataclass, field.name) is not UNDEFINED
    }
    return items


def to_snake_case(name: str) -> str:
    return _PATTERN.sub("_", name).lower()


def to_camel_case(name: str) -> str:
    return "".join(n.title() for n in name.split("_"))



from marshy.factory.impl_marshaller_factory import register_impl
from marshy.marshaller_context import MarshallerContext

from persisty.impl.sqlalchemy.search_filter.and_filter_converter import (
    AndFilterConverter,
)
from persisty.impl.sqlalchemy.search_filter.field_filter_converter import (
    AttrFilterConverter,
)
from persisty.impl.sqlalchemy.search_filter.include_all_converter import (
    IncludeAllConverter,
)
from persisty.impl.sqlalchemy.search_filter.not_filter_converter import (
    NotFilterConverter,
)
from persisty.impl.sqlalchemy.search_filter.or_filter_converter import OrFilterConverter
from persisty.impl.sqlalchemy.search_filter.query_filter_converter import (
    QueryFilterConverter,
)
from persisty.impl.sqlalchemy.search_filter.search_filter_converter_abc import (
    SearchFilterConverterABC,
)
from persisty.impl.sqlalchemy.sqlalchemy_context_factory import SqlalchemyContextFactory
from persisty.impl.sqlalchemy.sqlalchemy_context_factory_abc import (
    SqlalchemyContextFactoryABC,
)


def configure_converters(context: MarshallerContext):
    register_impl(SearchFilterConverterABC, AndFilterConverter, context)
    register_impl(SearchFilterConverterABC, AttrFilterConverter, context)
    register_impl(SearchFilterConverterABC, IncludeAllConverter, context)
    register_impl(SearchFilterConverterABC, NotFilterConverter, context)
    register_impl(SearchFilterConverterABC, OrFilterConverter, context)
    register_impl(SearchFilterConverterABC, QueryFilterConverter, context)


def configure_sqlalchemy_context(context: MarshallerContext):
    register_impl(SqlalchemyContextFactoryABC, SqlalchemyContextFactory, context)



from dataclasses import dataclass

from persisty.index.index_abc import IndexABC


@dataclass
class AttrIndex(IndexABC):
    """
    Standard single attribute index, usually translated into a btree and used for boosting performance
    """

    attr_name: str



from typing import Optional, Tuple, Any

from sqlalchemy import Table, or_

from persisty.impl.sqlalchemy.search_filter.and_filter_converter import (
    AndFilterConverter,
)
from persisty.search_filter.or_filter import Or
from persisty.search_filter.search_filter_abc import SearchFilterABC
from persisty.store_meta import StoreMeta


class OrFilterConverter(AndFilterConverter):
    def convert(
        self,
        search_filter: SearchFilterABC,
        table: Table,
        store_meta: StoreMeta,
        context,
    ) -> Optional[Tuple[Any, bool]]:
        if isinstance(search_filter, Or):
            sub_clauses, handled = self.get_sub_clauses(
                search_filter.search_filters, table, store_meta, context
            )
            clause = or_(*sub_clauses) if sub_clauses else None
            return clause, handled



from datetime import datetime, timezone

from persisty.attr.generator.attr_value_generator_abc import AttrValueGeneratorABC
from persisty.util.undefined import UNDEFINED


class TimestampGenerator(AttrValueGeneratorABC):
    def transform(self, value, item):
        if value is UNDEFINED:
            return datetime.now().astimezone(timezone.utc)
        return value



from typing import Dict

from servey.finder.action_finder_abc import find_actions_with_trigger_type
from servey.servey_celery.celery_config.celery_config_abc import CeleryConfigABC
from celery import Celery

from persisty.trigger.after_create_trigger import AfterCreateTrigger
from persisty.trigger.after_delete_trigger import AfterDeleteTrigger
from persisty.trigger.after_update_trigger import AfterUpdateTrigger


class CeleryStoreTriggerConfig(CeleryConfigABC):
    def configure(self, app: Celery, global_ns: Dict):
        _configure_for_trigger_type(app, global_ns, AfterCreateTrigger)
        _configure_for_trigger_type(app, global_ns, AfterUpdateTrigger)
        _configure_for_trigger_type(app, global_ns, AfterDeleteTrigger)


def _configure_for_trigger_type(app: Celery, global_ns: Dict, trigger_type):
    for _action, _ in find_actions_with_trigger_type(trigger_type):
        if _action.name not in global_ns:
            global_ns[_action.name] = app.task(_action.fn)



from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Optional, Generic, Iterator

from servey.action.action import Action
from servey.security.authorization import Authorization

from persisty.search_filter.search_filter_factory import search_filter_dataclass_for
from persisty.search_order.search_order_factory import search_order_dataclass_for
from persisty.store.store_abc import StoreABC, T
from persisty.store_meta import StoreMeta

ROUTE = "starlette.routing.Route"


@dataclass
class StoreFactoryABC(Generic[T], ABC):
    """
    Factory for store objects which allows defining access control rules for store.
    """

    @abstractmethod
    def get_meta(self) -> StoreMeta:
        """Get the meta for the store"""

    @abstractmethod
    def create(self, authorization: Optional[Authorization]) -> Optional[StoreABC[T]]:
        """Create a new store instance"""

    # pylint: disable=R0914
    def create_actions(self) -> Iterator[Action]:
        """Create actions for this factory"""
        from persisty.servey.actions import wrap_links_in_actions

        store_meta = self.get_meta()
        store_access = store_meta.store_access
        item_type = wrap_links_in_actions(store_meta.get_read_dataclass())
        search_filter_type = search_filter_dataclass_for(store_meta)
        search_order_type = search_order_dataclass_for(store_meta)
        create_input_type = store_meta.get_create_dataclass()
        update_input_type = store_meta.get_update_dataclass()

        if store_access.creatable:
            from persisty.servey.actions import action_for_create

            yield action_for_create(
                self,
                item_type,
                create_input_type,
            )
        if store_access.readable:
            from persisty.servey.actions import action_for_read

            yield action_for_read(self, item_type)
        if store_access.updatable:
            from persisty.servey.actions import action_for_update

            yield action_for_update(
                self,
                item_type,
                update_input_type,
                search_filter_type,
            )
        if store_access.deletable:
            from persisty.servey.actions import action_for_delete

            yield action_for_delete(self)
        if store_access.searchable:
            from persisty.servey.actions import action_for_search

            yield action_for_search(
                self,
                item_type,
                search_filter_type,
                search_order_type,
            )
            from persisty.servey.actions import action_for_count

            yield action_for_count(self, search_filter_type)
        if store_access.readable:
            from persisty.servey.actions import action_for_read_batch

            yield action_for_read_batch(self, item_type)
        if store_access.editable:
            from persisty.servey.actions import action_for_edit_batch

            yield action_for_edit_batch(
                self,
                create_input_type,
                update_input_type,
            )

    # noinspection PyMethodMayBeStatic
    def create_routes(self) -> Iterator[ROUTE]:
        """Create routes for this factory"""
        return iter(tuple())



from dataclasses import dataclass
from typing import Optional, Type

from servey.security.authorization import Authorization

from persisty.attr.attr_filter import AttrFilter, AttrFilterOp
from persisty.factory.store_factory_abc import StoreFactoryABC
from persisty.link.linked_store_abc import LinkedStoreABC
from persisty.search_filter.exclude_all import EXCLUDE_ALL
from persisty.search_filter.search_filter_abc import SearchFilterABC
from persisty.util import to_snake_case, UNDEFINED


@dataclass
class HasCountCallable:
    store_factory: StoreFactoryABC
    search_filter: SearchFilterABC

    def __call__(self, authorization: Optional[Authorization] = None) -> int:
        store = self.store_factory.create(authorization)
        count = store.count(search_filter=self.search_filter)
        return count


@dataclass
class HasCount(LinkedStoreABC):
    local_key_attr_name: str = "id"
    remote_key_attr_name: Optional[str] = None

    def __get__(self, obj, obj_type) -> HasCountCallable:
        key = getattr(obj, self.local_key_attr_name)
        if key is UNDEFINED:
            search_filter = EXCLUDE_ALL
        else:
            search_filter = AttrFilter(self.remote_key_attr_name, AttrFilterOp.eq, key)
        return HasCountCallable(
            store_factory=self.get_linked_store_factory(),
            search_filter=search_filter,
        )

    def __set_name__(self, owner, name):
        self.name = name
        if self.remote_key_attr_name is None:
            self.remote_key_attr_name = f"{to_snake_case(owner.__name__)}_id"

    def get_name(self) -> str:
        return self.name

    def get_linked_type(self, forward_ref_ns: str) -> Type[int]:
        return int



from typing import Optional, Tuple, Any

from sqlalchemy import Table, and_

from persisty.impl.sqlalchemy.search_filter.search_filter_converter_abc import (
    SearchFilterConverterABC,
)
from persisty.search_filter.and_filter import And
from persisty.search_filter.search_filter_abc import SearchFilterABC
from persisty.store_meta import StoreMeta


class AndFilterConverter(SearchFilterConverterABC):
    def convert(
        self,
        search_filter: SearchFilterABC,
        table: Table,
        store_meta: StoreMeta,
        context,
    ) -> Optional[Tuple[Any, bool]]:
        if isinstance(search_filter, And):
            sub_clauses, handled = self.get_sub_clauses(
                search_filter.search_filters, table, store_meta, context
            )
            clause = and_(*sub_clauses) if sub_clauses else None
            return clause, handled

    @staticmethod
    def get_sub_clauses(
        search_filters: Tuple[SearchFilterABC],
        table: Table,
        store_meta: StoreMeta,
        context,
    ) -> Optional[Tuple[Any, bool]]:
        sub_clauses = []
        handled = True
        for sub_filter in search_filters:
            sub_clause, sub_handled = context.convert(sub_filter, table, store_meta)
            if sub_clause is not None:
                sub_clauses.append(sub_clause)
            handled = handled and sub_handled
        return sub_clauses, handled



from dataclasses import dataclass, field
from sqlalchemy import MetaData, Table
from sqlalchemy.engine import Engine

from persisty.impl.sqlalchemy.sqlalchemy_table_converter import SqlalchemyTableConverter
from persisty.store_meta import StoreMeta


@dataclass
class SqlalchemyContext:
    engine: Engine
    developer_mode: bool = False
    meta_data: MetaData = field(default_factory=MetaData)

    @property
    def converter(self) -> SqlalchemyTableConverter:
        converter = getattr(self, "_converter", None)
        if not converter:
            converter = SqlalchemyTableConverter(self.engine, self.meta_data)
            setattr(self, "_converter", converter)
        return converter

    def get_table(self, store_meta: StoreMeta) -> Table:
        tables_by_store_name = getattr(self, "_tables_by_store_name", None)
        if not tables_by_store_name:
            tables_by_store_name = {}
            setattr(self, "_tables_by_store_name", tables_by_store_name)
        table = tables_by_store_name.get(store_meta.name)
        if table is None:
            table, indexes = self.converter.to_sqlalchemy_table_and_indexes(store_meta)
            tables_by_store_name[store_meta.name] = table
            if self.developer_mode:
                table.create(self.engine)
                for index in indexes:
                    index.create(self.engine)
        return table



import inspect
from dataclasses import dataclass, field
from typing import Optional, Type

from boto3.dynamodb.types import TypeDeserializer
from marshy import ExternalType, get_default_context
from marshy.marshaller.marshaller_abc import MarshallerABC
from marshy.marshaller_context import MarshallerContext
from marshy.types import ExternalItemType
from servey.action.action import Action
from servey.servey_aws.event_handler.event_handler_abc import (
    EventHandlerABC,
    EventHandlerFactoryABC,
)

from persisty.trigger.after_create_trigger import AfterCreateTrigger
from persisty.trigger.after_delete_trigger import AfterDeleteTrigger
from persisty.trigger.after_update_trigger import AfterUpdateTrigger


@dataclass
class DynamodbPostProcessEventHandler(EventHandlerABC):
    """
    Event handler for events in dynamodb format
    """

    action: Action
    item_marshaller: MarshallerABC

    def is_usable(self, event: ExternalItemType, context) -> bool:
        # noinspection PyBroadException
        try:
            # noinspection PyTypeChecker
            result = bool(event["Records"][0]["dynamodb"])
            return result
        except Exception:
            return False

    def handle(self, event: ExternalItemType, context) -> ExternalType:
        deserializer = TypeDeserializer()
        for record in event["Records"]:
            # noinspection PyTypeChecker
            new_image = record["NewImage"]
            if new_image:
                new_image = deserializer.deserialize(new_image)
                new_image = self.item_marshaller.load(new_image)
            # noinspection PyTypeChecker
            old_image = record["OldImage"]
            if old_image:
                old_image = deserializer.deserialize(old_image)
                old_image = self.item_marshaller.load(old_image)
            if old_image and new_image:
                if _has_trigger(self.action, AfterUpdateTrigger):
                    self.action.fn(old_image, new_image)
            elif new_image:
                if _has_trigger(self.action, AfterCreateTrigger):
                    self.action.fn(new_image)
            elif _has_trigger(self.action, AfterDeleteTrigger):
                self.action.fn(old_image)


@dataclass
class DynamodbPostProcessEventHandlerFactory(EventHandlerFactoryABC):
    marshaller_context: MarshallerContext = field(default_factory=get_default_context)

    def create(self, action: Action) -> Optional[EventHandlerABC]:
        if (
            _has_trigger(action, AfterCreateTrigger)
            or _has_trigger(action, AfterUpdateTrigger)
            or _has_trigger(action, AfterDeleteTrigger)
        ):
            sig = inspect.signature(action.fn)
            item_type = next(iter(sig.parameters.values())).annotation
            item_marshaller = self.marshaller_context.get_marshaller(item_type)
            return DynamodbPostProcessEventHandler(action, item_marshaller)


def _has_trigger(action_: Action, trigger_type: Type):
    result = next(
        (True for t in action_.triggers if isinstance(t, trigger_type)), False
    )
    return result



from abc import ABC, abstractmethod
from typing import Iterator, List, Optional, Tuple, Dict

from persisty.errors import PersistyError
from persisty.batch_edit import BatchEdit
from persisty.batch_edit_result import BatchEditResult
from persisty.result_set import ResultSet
from persisty.search_filter.include_all import INCLUDE_ALL
from persisty.search_filter.search_filter_abc import SearchFilterABC
from persisty.search_order.search_order import SearchOrder
from persisty.store.store_abc import StoreABC
from persisty.store_meta import StoreMeta
from persisty.store.wrapper_store_abc import WrapperStoreABC, T
from persisty.util import get_logger
from persisty.util.encrypt_at_rest import decrypt, encrypt

logger = get_logger(__name__)


# noinspection PyMethodMayBeStatic
class FilteredStoreABC(WrapperStoreABC[T], ABC):
    """Extension to wrapper storage which loads items before updates / deletes for the purpose of filtering"""

    @abstractmethod
    def get_store(self) -> StoreABC:
        """Get wrapped store"""

    def get_meta(self) -> StoreMeta:
        return self.get_store().get_meta()

    def filter_create(self, item: T) -> Optional[T]:
        """search_filter an stored before create"""
        return item

    # pylint: disable=W0613
    # noinspection PyUnusedLocal
    def filter_update(self, item: T, updates: T) -> T:
        """search_filter an stored before create"""
        return updates

    def filter_read(self, item: T) -> Optional[T]:
        """search_filter an stored after read"""
        return item

    # pylint: disable=W0613
    # noinspection PyUnusedLocal
    def allow_delete(self, item: T) -> bool:
        """Filter a delete of an stored"""
        return True

    def filter_search_filter(
        self, search_filter: SearchFilterABC
    ) -> Tuple[SearchFilterABC, bool]:
        """
        filter a search_filter. Return the result and a boolean indicating if the filter has been fully
        handled by this operation
        """
        return search_filter, True

    def create(self, item: T) -> Optional[T]:
        item = self.filter_create(item)
        if item:
            return self.get_store().create(item)

    def read(self, key: str) -> Optional[T]:
        item = self.get_store().read(key)
        if item:
            item = self.filter_read(item)
        return item

    def read_batch(self, keys: List[str]) -> List[Optional[T]]:
        assert len(keys) <= self.get_meta().batch_size
        items = self.get_store().read_batch(keys)
        items = [self.filter_read(item) if item else None for item in items]
        return items

    # pylint: disable=W0212
    def _update(self, key: str, item: T, updates: T) -> Optional[T]:
        updates = self.filter_update(item, updates)
        if not updates:
            return None
        return self.get_store()._update(key, item, updates)

    # pylint: disable=W0212
    def _delete(self, key: str, item: T) -> bool:
        if self.allow_delete(item):
            return self.get_store()._delete(key, item)
        return False

    def search(
        self,
        search_filter: SearchFilterABC = INCLUDE_ALL,
        search_order: Optional[SearchOrder] = None,
        page_key: Optional[str] = None,
        limit: Optional[int] = None,
    ) -> ResultSet[T]:
        if limit is None:
            limit = self.get_meta().batch_size
        else:
            assert limit <= self.get_meta().batch_size
        search_filter, fully_handled = self.filter_search_filter(search_filter)
        if fully_handled:
            return self.get_store().search(search_filter, search_order, page_key, limit)
        # Since the nested search_filter was not fully able to handle the constraint, we handle it here...
        nested_page_key = None
        nested_item_key = None
        if page_key:
            nested_page_key, nested_item_key = decrypt(page_key)
        key_config = self.get_meta().key_config
        results = []
        while True:
            result_set = self.get_store().search(
                search_filter, search_order, nested_page_key
            )
            items = (self.filter_read(item) for item in result_set.results)
            items = (item for item in items if item)  # Remove filtered items

            # Skip over any results we need to in the current page
            while nested_item_key:
                next_item = next(items, None)
                if not next_item:
                    raise PersistyError(
                        "invalid_page_key"
                    )  # The item was probably deleted
                next_item_key = key_config.to_key_str(next_item)
                if next_item_key == nested_item_key:
                    nested_item_key = None

            results.extend(items)
            if len(results) == limit:
                if not result_set.next_page_key:
                    return ResultSet(results)
                return ResultSet(results, encrypt([result_set.next_page_key, None]))
            if len(results) > limit:
                results = results[:limit]
                return ResultSet(
                    results,
                    encrypt([nested_page_key, key_config.to_key_str(results[-1])]),
                )
            if not result_set.next_page_key:
                return ResultSet(results)
            nested_page_key = result_set.next_page_key

    def search_all(
        self,
        search_filter: SearchFilterABC = INCLUDE_ALL,
        search_order: Optional[SearchOrder] = None,
    ) -> Iterator[T]:
        search_filter, fully_handled = self.filter_search_filter(search_filter)
        items = self.get_store().search_all(search_filter, search_order)
        if fully_handled:
            yield from items
            return
        for item in items:
            item = self.filter_read(item)
            if item:
                yield item

    def count(self, search_filter: SearchFilterABC = INCLUDE_ALL) -> int:
        search_filter, fully_handled = self.filter_search_filter(search_filter)
        if fully_handled:
            return self.get_store().count(search_filter)
        items = self.get_store().search_all(search_filter)
        items = (self.filter_read(item) for item in items)
        items = (item for item in items if item)  # Remove filtered items
        count = sum(1 for _ in items)
        return count

    # pylint: disable=R0912
    def _edit_batch(
        self, edits: List[BatchEdit[T, T]], items_by_key: Dict[str, T]
    ) -> List[BatchEditResult[T, T]]:
        assert len(edits) <= self.get_meta().batch_size
        key_config = self.get_meta().key_config

        results = [BatchEditResult[T, T](edit, code="unknown") for edit in edits]
        results_by_id = {result.edit.id: result for result in results}
        filtered_edits = []
        for edit in edits:
            result = results_by_id[edit.id]
            try:
                if edit.create_item:
                    item = self.filter_create(edit.create_item)
                    if not item:
                        result.code = "fitered_edit"
                        continue
                    filtered_edits.append(BatchEdit[T, T](create_item=item, id=edit.id))
                elif edit.update_item:
                    key = edit.get_key(key_config)
                    item = items_by_key.get(key)
                    if not item:
                        result.code = "item_missing"
                        continue
                    filtered_updates = self.filter_update(item, edit.update_item)
                    if not filtered_updates:
                        result.code = "filtered_edit"
                        continue
                    filtered_edits.append(
                        BatchEdit[T, T](update_item=filtered_updates, id=edit.id)
                    )
                else:
                    item = items_by_key.get(edit.delete_key)
                    if not item:
                        result.code = "item_missing"
                        continue
                    if not self.allow_delete(item):
                        result.code = "filtered_edit"
                        continue
                    filtered_edits.append(edit)
            except PersistyError as e:
                result.msg = str(e)
        if filtered_edits:
            filtered_results = self.get_store()._edit_batch(
                filtered_edits, items_by_key
            )
            for filtered_result in filtered_results:
                result = results_by_id.get(filtered_result.edit.id)
                # If result is missing, then the nested storage has not implemented the protocol correctly
                if result:
                    result.copy_from(filtered_result)
        return results



from __future__ import annotations
from dataclasses import dataclass, fields


@dataclass(frozen=True)
class StoreAccess:
    creatable: bool = True
    readable: bool = True
    updatable: bool = True
    deletable: bool = True
    searchable: bool = True

    def __and__(self, other: StoreAccess) -> StoreAccess:
        result = StoreAccess(
            **{
                f.name: getattr(self, f.name) and getattr(other, f.name)
                for f in fields(StoreAccess)
            }
        )
        return result

    def __or__(self, other: StoreAccess) -> StoreAccess:
        result = StoreAccess(
            **{
                f.name: getattr(self, f.name) or getattr(other, f.name)
                for f in fields(StoreAccess)
            }
        )
        return result

    @property
    def editable(self):
        return self.creatable or self.updatable or self.deletable


NO_ACCESS = StoreAccess(**{f.name: False for f in fields(StoreAccess)})
ALL_ACCESS = StoreAccess()
READ_ONLY = StoreAccess(creatable=False, updatable=False, deletable=False)



from dataclasses import dataclass
from typing import Optional, Iterator

from marshy.types import ExternalItemType

from persisty.key_config.attr_key_config import AttrKeyConfig
from persisty.key_config.key_config_abc import KeyConfigABC, T
from persisty.util.encrypt_at_rest import decrypt, encrypt


@dataclass(frozen=True)
class DynamodbKeyConfig(KeyConfigABC):
    pk_attr: AttrKeyConfig
    sk_attr: Optional[AttrKeyConfig] = None

    def to_key_str(self, item: T) -> str:
        if not self.sk_attr:
            return self.pk_attr.to_key_str(item)
        return encrypt([self.pk_attr.to_key_str(item), self.sk_attr.to_key_str(item)])

    def from_key_str(self, key: Optional[str], target: T):
        if self.sk_attr:
            key = decrypt(key)
            self.pk_attr.from_key_str(key[0], target)
            self.sk_attr.from_key_str(key[1], target)
        else:
            self.pk_attr.from_key_str(key, target)

    def to_key_dict(self, key: Optional[str]) -> ExternalItemType:
        if self.sk_attr:
            key = decrypt(key)
            return {
                self.pk_attr.attr_name: self.pk_attr.to_key_attr(key[0]),
                self.sk_attr.attr_name: self.sk_attr.to_key_attr(key[1]),
            }
        return self.pk_attr.to_key_dict(key)

    def get_key_attrs(self) -> Iterator[str]:
        yield self.pk_attr
        yield self.sk_attr
